"""ica contains routines for creating and working with
incremental capacity analysis data"""

import os
import logging
import warnings

import numpy as np
import pandas as pd
from scipy import stats
from scipy.interpolate import interp1d
from scipy.signal import savgol_filter
from scipy.integrate import simps
from scipy.ndimage.filters import gaussian_filter1d
import pandas as pd

from cellpy.exceptions import NullData
from cellpy.readers.core import collect_capacity_curves


# TODO: @jepe - documentation and tests
# TODO: @jepe - fitting of o-c curves and differentiation
# TODO: @jepe - modeling and fitting
# TODO: @jepe - full-cell
# TODO: @jepe - binning method (assigned to Asbjoern)


class Converter:
    """Class for dq-dv handling.

    Typical usage is to (1) set the data, (2) inspect the data,
    (3) pre-process the data,
    (4) perform the dq-dv transform, and finally (5) post-process the data.

    A short note about normalization:

        - If ``normalization`` is set to ``False``, then no normalization will be done.
        - If ``normalization`` is ``True``, and ``normalization_factor`` is ``None``, the total capacity of
          the half cycle will be used for normalization, else the ``normalization_factor`` will be used.
        - If ``normalization`` is ``True``, and ``normalization_roof`` is not ``None``,
          the capacity divided by ``normalization_roof`` will be used for normalization.

    """

    def __init__(
        self,
        capacity=None,
        voltage=None,
        points_pr_split=10,
        max_points=None,
        voltage_resolution=None,
        capacity_resolution=None,
        minimum_splits=3,
        interpolation_method="linear",
        increment_method="diff",
        pre_smoothing=False,
        smoothing=False,
        post_smoothing=True,
        normalize=True,
        normalizing_factor=None,
        normalizing_roof=None,
        savgol_filter_window_divisor_default=50,
        savgol_filter_window_order=3,
        voltage_fwhm=0.01,
        gaussian_order=0,
        gaussian_mode="reflect",
        gaussian_cval=0.0,
        gaussian_truncate=4.0,
    ):

        self.capacity = capacity
        self.voltage = voltage

        self.capacity_preprocessed = None
        self.voltage_preprocessed = None
        self.capacity_inverted = None
        self.voltage_inverted = None

        self.incremental_capacity = None
        self._incremental_capacity = None  # before smoothing
        self.voltage_processed = None
        self._voltage_processed = None  # before shifting / centering

        self.voltage_inverted_step = None

        self.points_pr_split = points_pr_split
        self.max_points = max_points
        self.voltage_resolution = voltage_resolution
        self.capacity_resolution = capacity_resolution
        self.minimum_splits = minimum_splits
        self.interpolation_method = interpolation_method
        self.increment_method = increment_method
        self.pre_smoothing = pre_smoothing
        self.smoothing = smoothing
        self.post_smoothing = post_smoothing
        self.savgol_filter_window_divisor_default = savgol_filter_window_divisor_default
        self.savgol_filter_window_order = savgol_filter_window_order
        self.voltage_fwhm = voltage_fwhm
        self.gaussian_order = gaussian_order
        self.gaussian_mode = gaussian_mode
        self.gaussian_cval = gaussian_cval
        self.gaussian_truncate = gaussian_truncate
        self.normalize = normalize
        self.normalizing_factor = normalizing_factor
        self.normalizing_roof = normalizing_roof

        self.d_capacity_mean = None
        self.d_voltage_mean = None
        self.len_capacity = None
        self.len_voltage = None
        self.min_capacity = None
        self.max_capacity = None
        self.start_capacity = None
        self.end_capacity = None
        self.number_of_points = None
        self.std_err_median = None
        self.std_err_mean = None

        self.fixed_voltage_range = False

        self.errors = []

    def __str__(self):
        txt = f"[ica.converter] {str(type(self))}\n"
        attrs = vars(self)
        for name, att in attrs.items():
            if isinstance(att, (pd.DataFrame, pd.Series, np.ndarray)):
                str_att = f"<vector> ({str(type(att))})"
            else:
                str_att = str(att)
            txt += f"{name}: {str_att}\n"

        return txt

    def set_data(self, capacity, voltage=None, capacity_label="q", voltage_label="v"):
        """Set the data."""

        logging.debug("setting data (capacity and voltage)")

        if isinstance(capacity, pd.DataFrame):
            logging.debug("received a pandas.DataFrame")
            self.capacity = capacity[capacity_label]
            self.voltage = capacity[voltage_label]
        else:
            assert len(capacity) == len(voltage)
            self.capacity = capacity
            self.voltage = voltage

    def inspect_data(self, capacity=None, voltage=None, err_est=False, diff_est=False):
        """Check and inspect the data."""

        logging.debug("inspecting the data")

        if capacity is None:
            capacity = self.capacity
        if voltage is None:
            voltage = self.voltage

        if capacity is None or voltage is None:
            raise NullData

        self.len_capacity = len(capacity)
        self.len_voltage = len(voltage)

        if self.len_capacity <= 1:
            raise NullData
        if self.len_voltage <= 1:
            raise NullData

        self.min_capacity, self.max_capacity = value_bounds(capacity)
        self.start_capacity, self.end_capacity = index_bounds(capacity)

        self.number_of_points = len(capacity)

        if diff_est:
            d_capacity = np.diff(capacity)
            d_voltage = np.diff(voltage)

            self.d_capacity_mean = np.mean(d_capacity)
            self.d_voltage_mean = np.mean(d_voltage)

        if err_est:

            splits = int(self.number_of_points / self.points_pr_split)
            rest = self.number_of_points % self.points_pr_split

            if splits < self.minimum_splits:
                txt = "no point in splitting, too little data"
                logging.debug(txt)
                self.errors.append("splitting: to few points")
            else:
                if rest > 0:
                    _cap = capacity[:-rest]
                    _vol = voltage[:-rest]
                else:
                    _cap = capacity
                    _vol = voltage

                c_pieces = np.split(_cap, splits)
                v_pieces = np.split(_vol, splits)
                # c_middle = int(np.amax(c_pieces) / 2)

                std_err = []
                c_pieces_avg = []
                for c, v in zip(c_pieces, v_pieces):
                    _slope, _intercept, _r_value, _p_value, _std_err = stats.linregress(
                        c, v
                    )
                    std_err.append(_std_err)
                    c_pieces_avg.append(np.mean(c))

                self.std_err_median = np.median(std_err)
                self.std_err_mean = np.mean(std_err)

        if not self.start_capacity == self.min_capacity:
            self.errors.append("capacity: start<>min")

        if not self.end_capacity == self.max_capacity:
            self.errors.append("capacity: end<>max")

        if self.normalizing_factor is None:
            self.normalizing_factor = self.end_capacity

        if self.normalizing_roof is not None:
            self.normalizing_factor = (
                self.normalizing_factor * self.end_capacity / self.normalizing_roof
            )

    def pre_process_data(self):
        """Perform some pre-processing of the data (i.e. interpolation)."""

        logging.debug("pre-processing the data")

        capacity = self.capacity
        voltage = self.voltage

        # performing an interpolation in v(q) space
        logging.debug(" - interpolating voltage(capacity)")
        c1, c2 = index_bounds(capacity)
        if self.max_points is not None:
            len_capacity = min(self.max_points, self.len_capacity)
        elif self.capacity_resolution is not None:
            len_capacity = int(round(abs(c2 - c1) / self.capacity_resolution, 0))
        else:
            len_capacity = self.len_capacity

        f = interp1d(capacity, voltage, kind=self.interpolation_method)

        self.capacity_preprocessed = np.linspace(c1, c2, len_capacity)
        self.voltage_preprocessed = f(self.capacity_preprocessed)

        if self.pre_smoothing:
            logging.debug(" - pre-smoothing (savgol filter window)")
            savgol_filter_window_divisor = np.amin(
                (self.savgol_filter_window_divisor_default, len_capacity / 5)
            )
            savgol_filter_window_length = int(
                len_capacity / savgol_filter_window_divisor
            )

            if savgol_filter_window_length % 2 == 0:
                savgol_filter_window_length -= 1
            savgol_filter_window_length = np.amax([3, savgol_filter_window_length])

            self.voltage_preprocessed = savgol_filter(
                self.voltage_preprocessed,
                savgol_filter_window_length,
                self.savgol_filter_window_order,
            )

    def increment_data(self):
        """Perform the dq-dv transform."""

        # NOTE TO ASBJOERN: Probably insert method for "binning" instead of
        # TODO: Asbjørn will insert "binning" here
        # differentiating here
        # (use self.increment_method as the variable for selecting method for)

        logging.debug("incrementing data")

        # ---- shifting to y-x ----------------------------------------
        v1, v2 = value_bounds(self.voltage_preprocessed)
        if self.voltage_resolution is not None:
            len_voltage = int(round(abs(v2 - v1) / self.voltage_resolution, 0))
        else:
            len_voltage = int(len(self.voltage_preprocessed))

        # ---- interpolating ------------------------------------------
        logging.debug(" - interpolating capacity(voltage)")
        f = interp1d(
            self.voltage_preprocessed,
            self.capacity_preprocessed,
            kind=self.interpolation_method,
        )

        self.voltage_inverted = np.linspace(v1, v2, len_voltage)
        self.voltage_inverted_step = (v2 - v1) / (len_voltage - 1)
        self.capacity_inverted = f(self.voltage_inverted)

        if self.smoothing:
            logging.debug(" - smoothing (savgol filter window)")
            savgol_filter_window_divisor = np.amin(
                (self.savgol_filter_window_divisor_default, len_voltage / 5)
            )

            savgol_filter_window_length = int(
                len(self.voltage_inverted) / savgol_filter_window_divisor
            )

            if savgol_filter_window_length % 2 == 0:
                savgol_filter_window_length -= 1

            self.capacity_inverted = savgol_filter(
                self.capacity_inverted,
                np.amax([3, savgol_filter_window_length]),
                self.savgol_filter_window_order,
            )

        # ---  diff --------------------
        if self.increment_method == "diff":
            logging.debug(" - diff using DIFF")
            self.incremental_capacity = (
                np.ediff1d(self.capacity_inverted) / self.voltage_inverted_step
            )
            self._incremental_capacity = self.incremental_capacity
            # --- need to adjust voltage ---
            self._voltage_processed = self.voltage_inverted[1:]
            self.voltage_processed = (
                self.voltage_inverted[1:] - 0.5 * self.voltage_inverted_step
            )

        elif self.increment_method == "hist":
            logging.debug(" - diff using HIST")
            # raise NotImplementedError

            df = pd.DataFrame(
                {"Capacity": self.capacity_inverted, "Voltage": self.voltage_inverted}
            )
            df["dQ"] = df.Capacity.diff()
            df["Voltage"] = df.Voltage.round(decimals=4)
            df = df.groupby(["Voltage"])["dQ"].sum().to_frame().reset_index()
            df["dV"] = df.Voltage.diff().rolling(1).sum()
            df["dQdV"] = df.dQ / df.dV
            # df = df[df.dQdV.notnull()]  # Might be needed, but could introduce an artefact

            self.incremental_capacity = df.dQdV
            self.voltage_processed = df.Voltage

            # TODO: Asbjoern, maybe you can put your method here? Yes

    def post_process_data(
        self, voltage=None, incremental_capacity=None, voltage_step=None
    ):
        """Perform post-processing (smoothing, normalisation, interpolation) of the data."""

        logging.debug("post-processing data")

        if voltage is None:
            voltage = self.voltage_processed
            incremental_capacity = self.incremental_capacity
            voltage_step = self.voltage_inverted_step

        if self.post_smoothing:
            logging.debug(" - post smoothing (gaussian)")
            logging.debug(f"    * using voltage fwhm: {self.voltage_fwhm}")
            points_fwhm = int(self.voltage_fwhm / voltage_step)

            sigma = np.amax([1, points_fwhm / 2])

            incremental_capacity = gaussian_filter1d(
                incremental_capacity,
                sigma=sigma,
                order=self.gaussian_order,
                mode=self.gaussian_mode,
                cval=self.gaussian_cval,
                truncate=self.gaussian_truncate,
            )

        if self.normalize:
            logging.debug(" - normalizing")
            area = simps(incremental_capacity, voltage)
            incremental_capacity = (
                incremental_capacity * self.normalizing_factor / abs(area)
            )

        self.incremental_capacity = incremental_capacity

        fixed_range = False
        if isinstance(self.fixed_voltage_range, np.ndarray):
            fixed_range = True
        else:
            if self.fixed_voltage_range:
                fixed_range = True
        if fixed_range:
            logging.debug(" - using fixed voltage range (interpolating)")
            v1, v2, number_of_points = self.fixed_voltage_range
            v = np.linspace(v1, v2, number_of_points)
            f = interp1d(
                x=self.voltage_processed,
                y=incremental_capacity,
                kind=self.interpolation_method,
                bounds_error=False,
                fill_value=np.NaN,
            )
            self.incremental_capacity = f(v)
            self.voltage_processed = v


def value_bounds(x):
    """Returns tuple with min and max in x."""
    return np.amin(x), np.amax(x)


def index_bounds(x):
    """Returns tuple with first and last item."""
    if isinstance(x, (pd.DataFrame, pd.Series)):
        return x.iloc[0], x.iloc[-1]
    else:
        return x[0], x[-1]


def dqdv_cycle(cycle, splitter=True, **kwargs):
    """Convenience functions for creating dq-dv data from given capacity and
    voltage cycle.

    Returns the DataFrame with a 'voltage' and a 'incremental_capacity'
    column.

    Args:
        cycle (pandas.DataFrame): the cycle data ('voltage', 'capacity', 'direction' (1 or -1)).
        splitter (bool): insert a np.NaN row between charge and discharge.

    Returns:
        List of step numbers corresponding to the selected steptype.
        Returns a ``pandas.DataFrame`` instead of a list if ``pdtype`` is set to ``True``.

    Additional key-word arguments are sent to Converter:

    Keyword Args:
        points_pr_split (int): only used when investigating data using splits, defaults to 10.
        max_points: None
        voltage_resolution (float): used for interpolating voltage data (e.g. 0.005)
        capacity_resolution: used for interpolating capacity data
        minimum_splits (int): defaults to 3.
        interpolation_method: scipy interpolation method
        increment_method (str): defaults to "diff"
        pre_smoothing (bool): set to True for pre-smoothing (window)
        smoothing (bool): set to True for smoothing during differentiation (window)
        post_smoothing (bool): set to True for post-smoothing (gaussian)
        normalize (bool): set to True for normalizing to capacity
        normalizing_factor (float):
        normalizing_roof  (float):
        savgol_filter_window_divisor_default (int): used for window smoothing, defaults to 50
        savgol_filter_window_order: used for window smoothing
        voltage_fwhm (float): used for setting the post-processing gaussian sigma, defaults to 0.01
        gaussian_order (int): defaults to 0
        gaussian_mode (str): defaults to "reflect"
        gaussian_cval (float): defaults to 0.0
        gaussian_truncate (float): defaults to 4.0

    Example:
        >>> cycle_df = my_data.get_cap(
        >>> ...   1,
        >>> ...   categorical_column=True,
        >>> ...   method = "forth-and-forth"
        >>> ...   insert_nan=False,
        >>> ... )
        >>> voltage, incremental = ica.dqdv_cycle(cycle_df)

    """

    if cycle.empty:
        raise NullData(f"The cycle (type={type(cycle)}) is empty.")

    c_first = cycle.loc[cycle["direction"] == -1]
    c_last = cycle.loc[cycle["direction"] == 1]

    converter = Converter(**kwargs)
    converter.set_data(c_first["capacity"], c_first["voltage"])
    converter.inspect_data()
    converter.pre_process_data()
    converter.increment_data()
    converter.post_process_data()
    voltage_first = converter.voltage_processed
    incremental_capacity_first = converter.incremental_capacity

    if splitter:
        voltage_first = np.append(voltage_first, np.NaN)
        incremental_capacity_first = np.append(incremental_capacity_first, np.NaN)

    converter = Converter(**kwargs)
    converter.set_data(c_last["capacity"], c_last["voltage"])
    converter.inspect_data()
    converter.pre_process_data()
    converter.increment_data()
    converter.post_process_data()
    voltage_last = converter.voltage_processed[::-1]
    incremental_capacity_last = converter.incremental_capacity[::-1]
    voltage = np.concatenate((voltage_first, voltage_last))
    incremental_capacity = np.concatenate(
        (incremental_capacity_first, incremental_capacity_last)
    )

    return voltage, incremental_capacity


def dqdv_cycles(cycles, not_merged=False, **kwargs):
    """Convenience functions for creating dq-dv data from given capacity and
    voltage cycles.

    Returns a DataFrame with a 'voltage' and a 'incremental_capacity'
    column.

    Args:
        cycles (pandas.DataFrame): the cycle data ('cycle', 'voltage',
             'capacity', 'direction' (1 or -1)).
        not_merged (bool): return list of frames instead of concatenating (
            defaults to False).

    Returns:
        ``pandas.DataFrame`` with columns 'cycle', 'voltage', 'dq'.

    Additional key-word arguments are sent to Converter:

    Keyword Args:
        points_pr_split (int): only used when investigating data using
            splits, defaults to 10.
        max_points: None
        voltage_resolution (float): used for interpolating voltage data
            (e.g. 0.005)
        capacity_resolution: used for interpolating capacity data
        minimum_splits (int): defaults to 3.
        interpolation_method: scipy interpolation method
        increment_method (str): defaults to "diff"
        pre_smoothing (bool): set to True for pre-smoothing (window)
        smoothing (bool): set to True for smoothing during
            differentiation (window)
        post_smoothing (bool): set to True for post-smoothing (gaussian)
        normalize (bool): set to True for normalizing to capacity
        normalizing_factor (float):
        normalizing_roof  (float):
        savgol_filter_window_divisor_default (int): used for window
            smoothing, defaults to 50
        savgol_filter_window_order: used for window smoothing
        voltage_fwhm (float): used for setting the post-processing
            gaussian sigma, defaults to 0.01
        gaussian_order (int): defaults to 0
        gaussian_mode (str): defaults to "reflect"
        gaussian_cval (float): defaults to 0.0
        gaussian_truncate (float): defaults to 4.0

    Example:
        >>> cycles_df = my_data.get_cap(
        >>> ...   categorical_column=True,
        >>> ...   method = "forth-and-forth",
        >>> ...   label_cycle_number=True,
        >>> ...   insert_nan=False,
        >>> ... )
        >>> ica_df = ica.dqdv_cycles(cycles_df)

    """

    # TODO: should add option for normalising based on first cycle capacity
    # this is e.g. done by first finding the first cycle capacity (nom_cap)
    # (or use nominal capacity given as input) and then propagating this to
    # Converter using the key-word arguments
    #   normalize=True, normalization_factor=1.0, normalization_roof=nom_cap

    ica_dfs = list()
    cycle_group = cycles.groupby("cycle")
    keys = list()
    for cycle_number, cycle in cycle_group:
        cycle = cycle.dropna()
        v, dq = dqdv_cycle(cycle, splitter=True, **kwargs)
        _ica_df = pd.DataFrame({"voltage": v, "dq": dq})
        if not not_merged:
            _ica_df["cycle"] = cycle_number
            _ica_df = _ica_df[["cycle", "voltage", "dq"]]
        else:
            keys.append(cycle_number)
            _ica_df = _ica_df[["voltage", "dq"]]
        ica_dfs.append(_ica_df)

    if not_merged:
        return keys, ica_dfs

    ica_df = pd.concat(ica_dfs)
    return ica_df


def dqdv(
    voltage,
    capacity,
    voltage_resolution=None,
    capacity_resolution=None,
    voltage_fwhm=0.01,
    pre_smoothing=True,
    diff_smoothing=False,
    post_smoothing=True,
    post_normalization=True,
    interpolation_method=None,
    gaussian_order=None,
    gaussian_mode=None,
    gaussian_cval=None,
    gaussian_truncate=None,
    points_pr_split=None,
    savgol_filter_window_divisor_default=None,
    savgol_filter_window_order=None,
    max_points=None,
    **kwargs,
):
    """Convenience functions for creating dq-dv data from given capacity
    and voltage data.

    Args:
        voltage: nd.array or pd.Series
        capacity: nd.array or pd.Series
        voltage_resolution: used for interpolating voltage data (e.g. 0.005)
        capacity_resolution: used for interpolating capacity data
        voltage_fwhm: used for setting the post-processing gaussian sigma
        pre_smoothing: set to True for pre-smoothing (window)
        diff_smoothing: set to True for smoothing during differentiation
            (window)
        post_smoothing: set to True for post-smoothing (gaussian)
        post_normalization: set to True for normalizing to capacity
        interpolation_method: scipy interpolation method
        gaussian_order: int
        gaussian_mode: mode
        gaussian_cval:
        gaussian_truncate:
        points_pr_split: only used when investigating data using splits
        savgol_filter_window_divisor_default: used for window smoothing
        savgol_filter_window_order: used for window smoothing
        max_points: restricting to max points in vector (capacity-selected)

    Returns:
        (voltage, dqdv)

    """
    # Notes:
    #     PEC data
    #         pre_smoothing = False
    #         diff_smoothing = False
    #         pos_smoothing = False
    #         voltage_resolution = 0.005
    #     PEC data
    #         ...
    #     Arbin data (IFE)
    #         ...

    converter = Converter(**kwargs)
    logging.debug("dqdv - starting")
    logging.debug("dqdv - created Converter obj")
    converter.pre_smoothing = pre_smoothing
    converter.post_smoothing = post_smoothing
    converter.smoothing = diff_smoothing
    converter.normalize = post_normalization
    converter.voltage_fwhm = voltage_fwhm
    logging.debug(f"converter.pre_smoothing: {converter.pre_smoothing}")
    logging.debug(f"converter.post_smoothing: {converter.post_smoothing}")
    logging.debug(f"converter.smoothing: {converter.smoothing}")
    logging.debug(f"converter.normalise: {converter.normalize}")
    logging.debug(f"converter.voltage_fwhm: {converter.voltage_fwhm}")

    if voltage_resolution is not None:
        converter.voltage_resolution = voltage_resolution

    if capacity_resolution is not None:
        converter.capacity_resolution = capacity_resolution

    if savgol_filter_window_divisor_default is not None:
        converter.savgol_filter_window_divisor_default = (
            savgol_filter_window_divisor_default
        )

        logging.debug(
            f"converter.savgol_filter_window_divisor_default: "
            f"{converter.savgol_filter_window_divisor_default}"
        )

    if savgol_filter_window_order is not None:
        converter.savgol_filter_window_order = savgol_filter_window_order

        logging.debug(
            f"converter.savgol_filter_window_order: "
            f"{converter.savgol_filter_window_order}"
        )

    if gaussian_mode is not None:
        converter.gaussian_mode = gaussian_mode

    if gaussian_order is not None:
        converter.gaussian_order = gaussian_order

    if gaussian_truncate is not None:
        converter.gaussian_truncate = gaussian_truncate

    if gaussian_cval is not None:
        converter.gaussian_cval = gaussian_cval

    if interpolation_method is not None:
        converter.interpolation_method = interpolation_method

    if points_pr_split is not None:
        converter.points_pr_split = points_pr_split

    if max_points is not None:
        converter.max_points = max_points

    converter.set_data(capacity, voltage)
    converter.inspect_data()
    converter.pre_process_data()
    converter.increment_data()
    converter.post_process_data()

    return converter.voltage_processed, converter.incremental_capacity


def dqdv_frames(cell, split=False, tidy=True, **kwargs):
    """Returns dqdv data as pandas.DataFrame(s) for all cycles.

    Args:
        cell (CellpyData-object).
        split (bool): return one frame for charge and one for
            discharge if True (defaults to False).
        tidy (bool): returns the split frames in wide format (defaults
            to True. Remark that this option is currently not available
            for non-split frames).

    Returns:
        one or two ``pandas.DataFrame`` with the following columns:
        cycle: cycle number (if split is set to True).
        voltage: voltage
        dq: the incremental capacity


    Additional key-word arguments are sent to Converter:

    Keyword Args:
        cycle (int or list of ints (cycle numbers)): will process all (or up to max_cycle_number)
            if not given or equal to None.
        points_pr_split (int): only used when investigating data
            using splits, defaults to 10.
        max_points: None
        voltage_resolution (float): used for interpolating voltage
            data (e.g. 0.005)
        capacity_resolution: used for interpolating capacity data
        minimum_splits (int): defaults to 3.
        interpolation_method: scipy interpolation method
        increment_method (str): defaults to "diff"
        pre_smoothing (bool): set to True for pre-smoothing (window)
        smoothing (bool): set to True for smoothing during
            differentiation (window)
        post_smoothing (bool): set to True for post-smoothing
            (gaussian)
        normalize (bool): set to True for normalizing to capacity
        normalizing_factor (float):
        normalizing_roof  (float):
        savgol_filter_window_divisor_default (int): used for window
            smoothing, defaults to 50
        savgol_filter_window_order: used for window smoothing
        voltage_fwhm (float): used for setting the post-processing
            gaussian sigma, defaults to 0.01
        gaussian_order (int): defaults to 0
        gaussian_mode (str): defaults to "reflect"
        gaussian_cval (float): defaults to 0.0
        gaussian_truncate (float): defaults to 4.0

    Example:
        >>> from cellpy.utils import ica
        >>> charge_df, dcharge_df = ica.ica_frames(my_cell, split=True)
        >>> charge_df.plot(x=("voltage", "v"))

    """
    # TODO: should add option for normalizing based on first cycle capacity
    # this is e.g. done by first finding the first cycle capacity (nom_cap)
    # (or use nominal capacity given as input) and then propagating this to
    # Converter using the key-word arguments
    #   normalize=True, normalization_factor=1.0, normalization_roof=nom_cap

    if split:
        return _dqdv_split_frames(cell, tidy=tidy, **kwargs)
    else:
        return _dqdv_combinded_frame(cell, tidy=tidy, **kwargs)


def _constrained_dq_dv_using_dataframes(capacity, minimum_v, maximum_v, **kwargs):
    converter = Converter(**kwargs)
    converter.set_data(capacity)
    converter.inspect_data()
    converter.pre_process_data()
    converter.increment_data()
    converter.fixed_voltage_range = [minimum_v, maximum_v, 100]
    converter.post_process_data()
    return converter.voltage_processed, converter.incremental_capacity


def _make_ica_charge_curves(cycles_dfs, cycle_numbers, minimum_v, maximum_v, **kwargs):
    incremental_charge_list = []

    for c, n in zip(cycles_dfs, cycle_numbers):
        if c.empty:
            logging.info(f"{n} is empty")
            v = [np.nan]
            dq = [np.nan]
        else:
            v, dq = _constrained_dq_dv_using_dataframes(
                c, minimum_v, maximum_v, **kwargs
            )
        if not incremental_charge_list:
            d = pd.DataFrame({"v": v})
            d.name = "voltage"
            incremental_charge_list.append(d)

            d = pd.DataFrame({f"dq": dq})
            d.name = n
            incremental_charge_list.append(d)

        else:
            d = pd.DataFrame({f"dq": dq})
            # d.name = f"{cycle}"
            d.name = n
            incremental_charge_list.append(d)

    return incremental_charge_list


def _dqdv_combinded_frame(cell, tidy=True, **kwargs):
    """Returns full cycle dqdv data for all cycles as one pd.DataFrame.

    Args:
        cell: CellpyData-object

    Returns:
        pandas.DataFrame with the following columns:
            cycle: cycle number
            voltage: voltage
            dq: the incremental capacity
    """
    cycle = kwargs.pop("cycle", None)
    cycles = cell.get_cap(
        cycle=cycle,
        method="forth-and-forth",
        categorical_column=True,
        label_cycle_number=True,
        insert_nan=False,
    )
    ica_df = dqdv_cycles(cycles, not_merged=not tidy, **kwargs)

    if not tidy:
        # dqdv_cycles returns a list of cycle numbers and a list of DataFrames
        # if not_merged is set to True (or not False)
        keys, ica_df = ica_df
        ica_df = pd.concat(ica_df, axis=1, keys=keys)
        return ica_df

    assert isinstance(ica_df, pd.DataFrame)
    return ica_df


def _dqdv_split_frames(
    cell,
    tidy=False,
    trim_taper_steps=None,
    steps_to_skip=None,
    steptable=None,
    max_cycle_number=None,
    **kwargs,
):
    """Returns dqdv data as pandas.DataFrames for all cycles.

    Args:
        cell (CellpyData-object).
        tidy (bool): return in wide format if False (default),
            long (tidy) format if True.

    Returns:
        (charge_ica_frame, discharge_ica_frame) where the frames are
        pandas.DataFrames where the first column is voltage ('v') and
        the following columns are the incremental capcaity for each
        cycle (multi-indexed, where cycle number is on the top level).

    Example:
        >>> from cellpy.utils import ica
        >>> charge_ica_df, dcharge_ica_df = ica.ica_frames(my_cell)
        >>> charge_ica_df.plot(x=("voltage", "v"))

    """
    cycle = kwargs.pop("cycle", None)
    if cycle and not isinstance(cycle, (list, tuple)):
        cycle = [cycle]

    charge_dfs, cycles, minimum_v, maximum_v = collect_capacity_curves(
        cell,
        direction="charge",
        trim_taper_steps=trim_taper_steps,
        steps_to_skip=steps_to_skip,
        steptable=steptable,
        max_cycle_number=max_cycle_number,
        cycle=cycle,
    )
    logging.debug(f"retrieved {len(charge_dfs)} charge cycles")
    # charge_df = pd.concat(
    # charge_dfs, axis=1, keys=[k.name for k in charge_dfs])

    ica_charge_dfs = _make_ica_charge_curves(
        charge_dfs, cycles, minimum_v, maximum_v, **kwargs
    )

    ica_charge_df = pd.concat(
        ica_charge_dfs, axis=1, keys=[k.name for k in ica_charge_dfs]
    )

    dcharge_dfs, cycles, minimum_v, maximum_v = collect_capacity_curves(
        cell,
        direction="discharge",
        trim_taper_steps=trim_taper_steps,
        steps_to_skip=steps_to_skip,
        steptable=steptable,
        max_cycle_number=max_cycle_number,
        cycle=cycle,
    )
    logging.debug(f"retrieved {len(dcharge_dfs)} discharge cycles")
    ica_dcharge_dfs = _make_ica_charge_curves(
        dcharge_dfs, cycles, minimum_v, maximum_v, **kwargs
    )
    ica_discharge_df = pd.concat(
        ica_dcharge_dfs, axis=1, keys=[k.name for k in ica_dcharge_dfs]
    )
    ica_charge_df.columns.names = ["cycle", "value"]
    ica_discharge_df.columns.names = ["cycle", "value"]

    if tidy:
        ica_charge_df = ica_charge_df.melt(
            "voltage", var_name="cycle", value_name="dq", col_level=0
        )
        ica_discharge_df = ica_discharge_df.melt(
            "voltage", var_name="cycle", value_name="dq", col_level=0
        )

    return ica_charge_df, ica_discharge_df


def _check_class_ica():
    print(40 * "=")
    print("running check_class_ica")
    print(40 * "-")

    import matplotlib.pyplot as plt

    cell = _get_a_cell_to_play_with()
    cycle = 5
    print("looking at cycle %i" % cycle)

    # ---------- processing and plotting ----------------
    fig, (ax1, ax2) = plt.subplots(2, 1)
    capacity, voltage = cell.get_ccap(cycle)
    ax1.plot(capacity, voltage, "b.-", label="raw")
    converter = Converter()
    converter.set_data(capacity, voltage)
    converter.inspect_data()
    converter.pre_process_data()
    ax1.plot(
        converter.capacity_preprocessed,
        converter.voltage_preprocessed,
        "r.-",
        alpha=0.3,
        label="pre-processed",
    )

    converter.increment_data()
    ax2.plot(
        converter.voltage_processed,
        converter.incremental_capacity,
        "b.-",
        label="incremented",
    )

    converter.fixed_voltage_range = False
    converter.post_smoothing = True
    converter.normalize = False
    converter.post_process_data()
    ax2.plot(
        converter.voltage_processed,
        converter.incremental_capacity,
        "y-",
        alpha=0.3,
        lw=4.0,
        label="smoothed",
    )

    converter.fixed_voltage_range = np.array((0.1, 1.2, 100))
    converter.post_smoothing = False
    converter.normalize = False
    converter.post_process_data()
    ax2.plot(
        converter.voltage_processed,
        converter.incremental_capacity,
        "go",
        alpha=0.7,
        label="fixed voltage range",
    )
    ax1.legend(numpoints=1)
    ax2.legend(numpoints=1)
    ax1.set_ylabel("Voltage (V)")
    ax1.set_xlabel("Capacity (mAh/g)")
    ax2.set_xlabel("Voltage (V)")
    ax2.set_ylabel("dQ/dV (mAh/g/V)")
    plt.show()


def _get_a_cell_to_play_with():
    from cellpy import cellreader

    # -------- defining overall path-names etc ----------
    current_file_path = os.path.dirname(os.path.realpath(__file__))
    print(current_file_path)
    relative_test_data_dir = "../../testdata/hdf5"
    test_data_dir = os.path.abspath(
        os.path.join(current_file_path, relative_test_data_dir)
    )
    # test_data_dir_out = os.path.join(test_data_dir, "out")
    test_cellpy_file = "20160805_test001_45_cc.h5"
    test_cellpy_file_full = os.path.join(test_data_dir, test_cellpy_file)
    # mass = 0.078609164

    # ---------- loading test-data ----------------------
    cell = cellreader.CellpyData()
    cell.load(test_cellpy_file_full)
    list_of_cycles = cell.get_cycle_numbers()
    number_of_cycles = len(list_of_cycles)
    print("you have %i cycles" % number_of_cycles)
    # cell.save(test_cellpy_file_full)
    return cell


def _check_if_works():
    import pandas as pd
    from cellpy import cellreader

    cell = _get_a_cell_to_play_with()

    a = dqdv_frames(cell)
    print("Hei")


if __name__ == "__main__":
    _check_if_works()



# -*- coding: utf-8 -*-
"""easyplot module for cellpy. It provides easy plotting of any cellpy-readable data using matplotlib.
Author: Amund M. Raniseth
Date: 01.07.2021
"""


import logging
import os
import warnings
from pathlib import Path
from re import S

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import lines
from matplotlib.artist import kwdoc
from matplotlib.lines import Line2D
from matplotlib.scale import LogScale
from matplotlib.ticker import FuncFormatter

import cellpy
from cellpy import log
from cellpy.parameters.internal_settings import (
    get_headers_journal,
    keys_journal_session,
)
from cellpy.utils.batch_tools.batch_journals import LabJournal

hdr_journal = get_headers_journal()

# Dictionary of all possible user input arguments(as keys) with example values of correct type
# Value is a tuple (immutable) of type and default value.
USER_PARAMS = {
    "cyclelife_plot": (bool, True),
    "cyclelife_separate_data": (
        bool,
        False,
    ),  # will plot each cyclelife datafile in separate plots
    "cyclelife_percentage": (bool, False),
    "cyclelife_coulombic_efficiency": (bool, False),
    "cyclelife_coulombic_efficiency_ylabel": (str, "Coulombic efficiency [%]"),
    "cyclelife_charge_c_rate": (bool, False),
    "cyclelife_discharge_c_rate": (bool, False),
    "cyclelife_c_rate_ylabel": (str, "Effective C-rate"),
    "cyclelife_ir": (bool, False),  # Allows user to plot IR data aswell
    "cyclelife_xlabel": (str, "Cycles"),
    "cyclelife_ylabel": (str, r"Capacity $\left[\frac{mAh}{g}\right]$"),
    "cyclelife_ylabel_percent": (str, "Capacity retention [%]"),
    "cyclelife_legend_outside": (
        bool,
        False,
    ),  # if True, the legend is placed outside the plot
    "cyclelife_degradation_slope": (
        bool,
        False,
    ),  # Adds simple degradation slope regression to plot
    "capacity_determination_from_ratecap": (
        bool,
        False,
    ),  # If True, uses the ratecap and capacity to determine the exp capacity
    "galvanostatic_plot": (bool, True),
    "galvanostatic_potlim": (tuple, None),  # min and max limit on potential-axis
    "galvanostatic_caplim": (tuple, None),
    "galvanostatic_xlabel": (str, r"Capacity $\left[\frac{mAh}{g}\right]$"),
    "galvanostatic_ylabel": (str, "Cell potential [V]"),
    "galvanostatic_normalize_capacity": (
        bool,
        False,
    ),  # Normalizes all cycles' capacity to 1.
    "dqdv_plot": (bool, False),
    "dqdv_potlim": (tuple, None),  # min and max limit on potential-axis
    "dqdv_dqlim": (tuple, None),
    "dqdv_xlabel": (
        str,
        r"dQ/dV $\left[\frac{mAh}{gV}\right]$",
    ),  # TODO what unit? jees
    "dqdv_ylabel": (str, "Cell potential [V]"),
    "specific_cycles": (list, None),
    "exclude_cycles": (list, None),
    "all_in_one": (
        bool,
        False,
    ),  # Decides if everything should be plotted in the same plot in GC and dQdV plot
    "only_dischg": (bool, False),  # Only show discharge curves
    "only_chg": (bool, False),  # Only show charge curves
    "outpath": (str, "./"),
    "outtype": (str, ".png"),  # What file format to save in
    "outname": (str, None),  # Overrides the automatic filename generation
    "figsize": (tuple, (6, 4)),  # 6 inches wide, 4 inches tall
    "figres": (int, 100),  # Dots per Inch
    "figtitle": (str, "Title"),  # None = original filepath
    "save_figures": (bool, True),
    "save_journal": (bool, False),  # Save journal
}


def help():
    """Method of the EasyPlot class which prints some helptext in addition to all supported params."""
    ## Prints out help page of this module
    help_str = (
        "The easyplot extension to cellpy aims to easily plot data in a pretty manner.\n"
        "In order to use this function, you must import cellpy, and easyplot from cellpy.utils.\n"
        "\n"
        "Usage:\n"
        "Create list of datafiles you want to plot on the following format:\n"
        "\n"
        "files = [\n"
        "\t'./folder/filename.ext',\n"
        "\t'./folder/filename2.ext',\n"
        "\t]\n"
        "\n"
        "And then call the easyplot.plot function with the files list as the first parameter, and any optional keyword arguments.\n"
        "Here is an example of the use of all keyword arguments:\n"
    )
    for kw in USER_PARAMS:
        if type(USER_PARAMS[kw][1]) == str:
            insert = "'" + USER_PARAMS[kw][1] + "'"
        else:
            insert = str(USER_PARAMS[kw][1])
        help_str += "\t" + kw + " = " + insert + ",\n"
    print(help_str)


class EasyPlot:
    """Main easyplot class.
    Takes all the inputs from the user in its kwargs upon object initialization.
    Gathers data, handles and plots it when object.plot() is called.

    Help: type easyplot.help()
    """

    def __init__(self, files=None, nicknames=None, journal=None, **kwargs):
        """Initialization function of the EasyPlot class.
        Input parameters:
        filenames (list of strings).
        nicknames (list of strings), must match length of filenames.
        journal (str or pathlib.Path object): journal file name (should not be used if files is given).
        any kwargs: use easyplot.help() to print all kwargs to terminal.

        Returns:
        easyplot object

        Most basic usage:
        ezpltobj = easyplot.EasyPlot(["name1", "name2"], None)"""

        # Make all user input variables of self
        self.files = files
        self.nicknames = nicknames
        self.kwargs = kwargs

        # More needed variables
        self.figs = []
        self.file_data = []
        self.use_arbin_sql = False
        if journal is not None:
            self.journal_file = Path(journal)
        else:
            self.journal_file = None
        self.journal = None

        # Dictionary of all possible user input arguments(as keys) with example values of correct type
        # Value is a tuple (immutable) of type and default value.
        self.user_params = USER_PARAMS

        # Create 'empty' attributes for later use
        self.outpath = None
        self.masses = None
        self.labels = None
        self.nom_caps = None
        self.colors = None

        # List of available colors

        # Fill in the rest of the variables from self.user_params if the user didn't specify
        self.fill_input()

        # Verify that the user input is sufficient
        self.verify_input()
        self._generate_list_of_available_colors()

    def _generate_list_of_available_colors(self):
        if 19 >= len(self.files) > 10:
            self.colors = [
                "#e6194b",
                "#3cb44b",
                "#ffe119",
                "#4363d8",
                "#f58231",
                "#911eb4",
                "#46f0f0",
                "#f032e6",
                "#bcf60c",
                "#fabebe",
                "#008080",
                "#e6beff",
                "#9a6324",
                "#fffac8",
                "#800000",
                "#aaffc3",
                "#808000",
                "#ffd8b1",
                "#000075",
                "#808080",
                "#000000",
            ]
            warnings.warn(
                "You inserted more than 10 datafiles! In a desperate attempt to keep "
                "the plots tidy, another colorpalette with 19 distinct colors were chosen."
            )
        elif len(self.files) > 19:
            warnings.warn(
                "You inserted more than 19 datafiles! We do not have that "
                "many colors in the palette, this some colors are beeing recycled. "
                "Keep track of the filenames and legends and make sure this doesn't confuse you."
            )
        else:
            self.colors = [
                "tab:blue",
                "tab:orange",
                "tab:green",
                "tab:red",
                "tab:purple",
                "tab:brown",
                "tab:pink",
                "tab:gray",
                "tab:olive",
                "tab:cyan",
            ] * 5

    def plot(self):
        """This is the method the user calls on his/hers easyplot object in order to gather the data and plot it.
        Usage: object.plot()"""

        # Load all cellpy files
        logging.debug("starting plotting")
        for file in self.files:
            if isinstance(file, (list, tuple)):
                logging.debug("linked files provided - need to merge")
                linked_files = True
            else:
                linked_files = False

            # If using arbin sql
            if self.use_arbin_sql:
                cpobj = cellpy.get(
                    filename=file, instrument="arbin_sql"
                )  # Initiate cellpy object
            else:  # Not Arbin SQL? Then its probably a local file

                # Check that file(s) exist
                if linked_files:
                    file_name = "_".join(file)
                    for _f in file:
                        if not os.path.isfile(_f):
                            logging.error("File not found: " + str(_f))
                            raise FileNotFoundError

                else:
                    file_name = file
                    if not os.path.isfile(file):
                        logging.error("File not found: " + str(file))
                        print(os.getcwd())
                        raise FileNotFoundError

                cpobj = cellpy.get(filename=file)  # Load regular file
                # Check that we get data
            if cpobj is None:
                warnings.warn(
                    f"File reader returned no data for filename {file}. Please make sure that the file exists or "
                    f"that the data exists in an eventual database."
                )

            # Get ID of all cycles
            cyc_nums = cpobj.get_cycle_numbers()

            # Only get the cycles which both exist in data, and that the user want
            if self.kwargs["specific_cycles"] is not None:
                cyc_not_available = (
                    set(cyc_nums) ^ set(self.kwargs["specific_cycles"])
                ) & set(self.kwargs["specific_cycles"])
                if len(cyc_not_available) > 0:
                    warn_str = (
                        f"You want to plot cycles which are not available in the data! Datafile(s): "
                        f"{file}"
                        f", Cycle(s): {str(cyc_not_available)}"
                    )
                    warnings.warn(warn_str)
                cyc_nums = list(
                    set(cyc_nums).intersection(self.kwargs["specific_cycles"])
                )

            if self.kwargs["exclude_cycles"] is not None:
                cyc_nums = list(set(cyc_nums) - set(self.kwargs["exclude_cycles"]))

            color = self.give_color()  # Get a color for the data

            self.file_data.append((cpobj, cyc_nums, color, file_name))

        # Check kwargs/input parameters to see what plots to make
        if self.kwargs["cyclelife_plot"]:
            self.plot_cyclelife()

        if self.kwargs["galvanostatic_plot"] and not self.kwargs["dqdv_plot"]:
            self.plot_gc()

        if self.kwargs["dqdv_plot"] and not self.kwargs["galvanostatic_plot"]:
            self.plot_dQdV()

        if self.kwargs["galvanostatic_plot"] and self.kwargs["dqdv_plot"]:
            self.plot_gc_and_dQdV()

        if self.kwargs["capacity_determination_from_ratecap"]:
            self.plot_cap_from_rc()

        self._wrap_up()

    def _wrap_up(self):
        # saving journal file
        if self.kwargs["save_journal"]:
            if self.journal is not None:
                if self.outpath is not None:
                    journal_file_path = Path(self.outpath) / self.journal_file.name
                else:
                    journal_file_path = self.journal_file.name

                # if we want to enforce that the file will be a xlsx file:
                # journal_file_path = journal_file_path.with_suffix(".xlsx")

                journal_file_path = journal_file_path.with_suffix(".json")
                self.journal.to_file(
                    file_name=journal_file_path, paginate=False, to_project_folder=False
                )
                xlsx_journal_file_path = journal_file_path.with_name(
                    f"{journal_file_path.stem}.xlsx"
                )
                self.journal.to_file(
                    file_name=xlsx_journal_file_path,
                    paginate=False,
                    to_project_folder=False,
                )

    def verify_input(self):
        """Verifies that the users' input to the object is correct."""
        # Check that output dir exist (or create one)
        self.outpath = self.handle_outpath()  # Takes care of the output path

        # Check the nicknames
        if self.nicknames:
            if len(self.nicknames) != len(self.files):
                logging.error(
                    "Use nicknames = None, or specify exactly one nickname per datafile. You have specified "
                    + str(len(self.nicknames))
                    + " nicknames while inputting "
                    + str(len(self.files))
                    + " datafiles"
                )
                raise AssertionError

        # Check that all kwargs are used correctly
        for key in self.kwargs:
            # Check that input parameter exist
            try:
                self.user_params[key]
            except KeyError as e:
                logging.error(
                    "Input parameter "
                    + key
                    + " is not a valid parameter! Please see example configuration for help or run easyplot.help()"
                )

            # Check that the type is correct
            if type(self.kwargs[key]) != self.user_params[key][0] and type(
                self.kwargs[key]
            ) != type(None):
                logging.error(
                    "Type of inputparameter for keyword '"
                    + key
                    + "' is wrong. The user specified "
                    + str(type(self.kwargs[key]))
                    + " but the program needs a "
                    + str(self.user_params[key][0])
                )
                raise TypeError

        # Check that the user isn't trying to plot "only" both discharge and charge.
        if self.kwargs["only_dischg"] and self.kwargs["only_chg"]:
            logging.error(
                "You can't plot 'only' discharge AND charge curves! Set one to False please."
            )

        if self.journal_file is not None:
            # Check that the user isn't providing both a list of files and a journal filename
            if self.files is not None:
                logging.error(
                    "You can't give both filenames and a journal file at the same time."
                )
                logging.error("Chose either filenames OR journal file name please.")
                raise ValueError
            self._read_journal_file()
            self._populate_from_journal()  # Temporary fix - the parameters should be read directly from journal later
        else:
            if self.files is None:
                logging.error("No file names provided.")
                logging.error("Add file names OR journal file name please.")
                raise ValueError

    def _read_journal_file(self):
        logging.debug(f"reading journal file {self.journal_file}")
        journal = LabJournal(db_reader=None)
        journal.from_file(self.journal_file, paginate=False)
        self.journal = journal

    def _populate_from_journal(self):
        logging.debug(f"populating from journal")
        # populating from only a subset of the available journal columns
        # - can be increased later
        try:
            self.files = self.journal.pages[hdr_journal["raw_file_names"]].to_list()
        except AttributeError:
            logging.debug("No raw files found in your journal")

        try:
            self.masses = self.journal.pages[hdr_journal["mass"]].to_list()
        except AttributeError:
            logging.debug("No masses found in your journal")

        try:
            self.labels = self.journal.pages[hdr_journal["label"]].to_list()
        except AttributeError:
            logging.debug("No labels found in your journal")

        try:
            self.nom_cap = self.journal.pages[hdr_journal["nom_cap"]].to_list()
        except AttributeError:
            logging.debug("No nominal capacity found in your journal")

        try:
            self.cellpy_files = self.journal.pages[
                hdr_journal["cellpy_file_name"]
            ].to_list()
        except AttributeError:
            logging.debug("No cellpy files found in your journal")

    def fill_input(self):
        """Fill in the rest of the variables from self.user_params if the user didn't specify"""
        # Can't just join dicts since they have differing formats, need to loop...
        for key in self.user_params:
            try:
                self.kwargs[key]
            except KeyError:
                self.kwargs[key] = self.user_params[key][1]

    def set_arbin_sql_credentials(
        self,
        server="localhost",
        uid="sa",
        pwd="Changeme123",
        driver="ODBC Driver 17 for SQL Server",
    ):
        """Sets cellpy.prms.Instruments.Arbin details to fit what is inserted.
        Parameters: Server = 'IP of server', uid = 'username', pwd = 'password', driver = 'ODBC Driver 17 for SQL Server'"""
        cellpy.prms.Instruments.Arbin["SQL_server"] = server
        cellpy.prms.Instruments.Arbin["SQL_UID"] = uid
        cellpy.prms.Instruments.Arbin["SQL_PWD"] = pwd
        cellpy.prms.Instruments.Arbin["SQL_Driver"] = driver
        self.use_arbin_sql = True

    def give_color(self):
        """Picks the first color from the color list and gives it away"""
        color = self.colors[0]
        self.colors = self.colors[1:]
        return color

    def give_fig(self):
        """Gives figure to whoever asks and appends it to figure list"""

        fig, ax = plt.subplots(figsize=(6, 4))
        self.figs.append((fig, ax))
        return fig, ax

    def handle_outpath(self):
        """Makes sure that self.outpath exists, or creates it."""
        out_path = self.kwargs["outpath"]

        # should make this a pathlib.Path object - but not sure if str is assumed later on in the code
        if os.path.isdir(out_path):
            logging.debug(f"out path set to {out_path}")
            return out_path
        elif not os.path.isdir(out_path):
            logging.debug(f"outpath does not exits - creating")
            try:
                os.makedirs(out_path)
                logging.debug(f"out path set to {out_path}")
                return out_path
            except OSError as e:
                logging.error(
                    f"Cannot create output directory {out_path}. Please make sure you "
                    f"have write permission. Error message: {e}"
                )

    def plot_cyclelife(self):
        """Takes all the parameters inserted in the object creation and plots cyclelife"""
        # Spawn fig and axis for plotting
        if not self.kwargs["cyclelife_separate_data"]:
            fig, ax = self.give_fig()
            if self.kwargs["cyclelife_coulombic_efficiency"]:
                # Spawn twinx axis and set label
                ax_ce = ax.twinx()
                ax_ce.set(ylabel=self.kwargs["cyclelife_coulombic_efficiency_ylabel"])
            if (
                self.kwargs["cyclelife_charge_c_rate"]
                or self.kwargs["cyclelife_discharge_c_rate"]
            ):
                ax_c_rate = ax.twinx()

                def format_label(x, pos):
                    # The commented out code here makes the fractioned C-rate like C/50 and so on.
                    """
                    if x >= 1:
                        s = '%.2gC' % x
                    elif x == 0:
                        s = r'C/$\infty$'
                    else:
                        newfloat = 1/x
                        s = 'C/%.2g' % newfloat
                    """
                    # The following just has decimal place C-rate.
                    s = "%.3gC" % x
                    return s

                ax_c_rate.yaxis.set_major_formatter(FuncFormatter(format_label))
                ax_c_rate.set(ylabel="Effective C-rate")

            if self.kwargs["cyclelife_ir"]:
                ax_ir = ax.twinx()

            outpath = self.outpath

        for cpobj, cyc_nums, color, filename in self.file_data:
            if self.kwargs["cyclelife_separate_data"]:
                fig, ax = self.give_fig()
                if self.kwargs["cyclelife_coulombic_efficiency"]:
                    # Spawn twinx axis and set label
                    ax_ce = ax.twinx()
                    ax_ce.set(
                        ylabel=self.kwargs["cyclelife_coulombic_efficiency_ylabel"]
                    )
                if (
                    self.kwargs["cyclelife_charge_c_rate"]
                    or self.kwargs["cyclelife_discharge_c_rate"]
                ):
                    ax_c_rate = ax.twinx()

                    def format_label(x, pos):
                        # The following just has decimal place C-rate.
                        s = "%.3gC" % x
                        return s

                    ax_c_rate.yaxis.set_major_formatter(FuncFormatter(format_label))
                    ax_c_rate.set(ylabel="Effective C-rate")

                if self.kwargs["cyclelife_ir"]:
                    ax_ir = ax.twinx()

            # Get Pandas DataFrame of pot vs cap from cellpy object
            df = cpobj.get_cap(
                method="forth-and-forth",
                label_cycle_number=True,
                categorical_column=True,
            )
            outpath += os.path.basename(filename).split(".")[0] + "_"

            # Group by cycle and make list of cycle numbers
            cycgrouped = df.groupby("cycle")
            keys = []
            for key, item in cycgrouped:
                keys.append(key)

            chgs = [[], []]  # List with cycle num and capacity
            dchgs = [[], []]
            # Accumulate cycles
            for cyc in keys:  # Loop over all cycles
                if cyc in cyc_nums:  # Check if it is in list of wanted cycles
                    cyc_df = cycgrouped.get_group(
                        cyc
                    )  # Get the group of datapoints from specific cycle

                    cyc_redox_grouped = cyc_df.groupby(
                        "direction"
                    )  # Group by direction (meaning if it is charging or discharging)

                    dchg_df = cyc_redox_grouped.get_group(
                        -1
                    )  # Data for the discharge curve
                    dchgs[0].append(cyc)  # Append to dchg list
                    dchgs[1].append(dchg_df["capacity"].iat[-2])

                    chg_df = cyc_redox_grouped.get_group(1)  # Data for the charge curve
                    chgs[0].append(cyc)  # Append to chg list
                    chgs[1].append(chg_df["capacity"].iat[-2])

            if self.kwargs[
                "cyclelife_percentage"
            ]:  # Normalize all datapoints on the first one
                norm_fact = (
                    dchgs[1][0] / 100
                )  # /100 is to get range from 0-100(%) in stead of 0-1
                for i in range(len(chgs[1])):
                    chgs[1][i] /= norm_fact
                for i in range(len(dchgs[1])):
                    dchgs[1][i] /= norm_fact

            # Make label from filename or nickname
            if self.nicknames:
                label = self.nicknames[self.files.index(filename)]
            else:
                label = str(os.path.basename(filename))
            # print("Discharge capacities:")
            # print(dchgs[1])
            # Actully place it in plot
            if not self.kwargs["only_dischg"] and not self.kwargs["only_chg"]:
                ax.scatter(chgs[0], chgs[1], c=color, alpha=0.2)
                ax.scatter(dchgs[0], dchgs[1], c=color, label=label)
            elif self.kwargs["only_dischg"]:
                ax.scatter(dchgs[0], dchgs[1], c=color, label=label)
            elif self.kwargs["only_chg"]:
                ax.scatter(chgs[0], chgs[1], c=color, alpha=0.2)

            if self.kwargs["cyclelife_coulombic_efficiency"]:
                # Get CE for cyc_nums
                coulombic_efficiency = cpobj.cell.summary[
                    "coulombic_efficiency_u_percentage"
                ]
                cycs = []
                CEs = []
                for cyc in keys:
                    if cyc in cyc_nums:
                        cycs.append(cyc)
                        CEs.append(coulombic_efficiency[cyc])

                # Place it in the plot
                ax_ce.scatter(cycs, CEs, c=color, marker="+")
                # print(filename + " Dchg 1-3: " + str(dchgs[1][0:3])  + ", CE 1-3: " + str(coulombic_efficiency[0:3]))

            if (
                self.kwargs["cyclelife_charge_c_rate"]
                or self.kwargs["cyclelife_discharge_c_rate"]
            ):
                # charge_c_rate = cpobj.cell.summary["charge_c_rate"] #This gives incorrect c-rates.

                stepstable = cpobj.cell.steps
                chg_c_rates, dchg_c_rates = get_effective_C_rates(stepstable)

                selected_chg_c_rates = []
                selected_dchg_c_rates = []
                selected_cycs = []

                for cyc in keys:
                    if cyc in cyc_nums:
                        selected_chg_c_rates.append(chg_c_rates[cyc - 1])
                        selected_dchg_c_rates.append(dchg_c_rates[cyc - 1])
                        selected_cycs.append(cyc)

                if (
                    self.kwargs["cyclelife_charge_c_rate"]
                    and not self.kwargs["cyclelife_discharge_c_rate"]
                ):
                    ax_c_rate.scatter(
                        selected_cycs, selected_chg_c_rates, c=color, marker="_"
                    )
                elif (
                    not self.kwargs["cyclelife_charge_c_rate"]
                    and self.kwargs["cyclelife_discharge_c_rate"]
                ):
                    ax_c_rate.scatter(
                        selected_cycs, selected_dchg_c_rates, c=color, marker="_"
                    )
                elif (
                    self.kwargs["cyclelife_charge_c_rate"]
                    and self.kwargs["cyclelife_discharge_c_rate"]
                ):
                    ax_c_rate.scatter(
                        selected_cycs, selected_chg_c_rates, c=color, marker="_"
                    )
                    ax_c_rate.scatter(
                        selected_cycs,
                        selected_dchg_c_rates,
                        c=color,
                        alpha=0.2,
                        marker="_",
                    )

            if self.kwargs["cyclelife_degradation_slope"]:
                from scipy.stats import linregress

                slope, intercept, r, p, se = linregress(dchgs[0], dchgs[1])
                x = np.linspace(0, ax.get_xlim()[1] * 0.9, 10)
                degradation_unit = (
                    r" $\frac{mAh}{g\cdot cycle}$"
                    if not self.kwargs["cyclelife_percentage"]
                    else r" $\frac{\%}{cycle}$"
                )
                intercept_unit = (
                    r" $\frac{mAh}{g}$"
                    if not self.kwargs["cyclelife_percentage"]
                    else r"%"
                )
                ax.plot(
                    x,
                    x * slope + intercept,
                    c=color,
                    label="Degradation: %g" % slope
                    + degradation_unit
                    + "\nIntercept:       %g" % intercept
                    + intercept_unit
                    + ", r=%g" % r,
                )

            """if self.kwargs["cyclelife_ir"]:
                chg_ir = []
                dchg_ir = []


                steptable = cpobj.steps
                print(steptable)
                newdf = steptable[["ir", "cycle", "type"]]
                for i,elem in enumerate(newdf.iterrows()):
                    if elem[1]["type"] == "charge":
                        chg_ir.append(elem[1]["ir"])
                    elif elem[1]["type"] == "discharge":
                        dchg_ir.append(elem[1]["ir"])
                print(chg_ir)

                for cyc in keys:
                    if cyc in cyc_nums:
                        ax_ir.scatter(cyc, chg_ir[cyc], c = color, marker = "*")
                        """

            if self.kwargs["cyclelife_separate_data"]:
                # Set all plot settings from Plot object
                self.fix_cyclelife(fig, ax)

                # Save fig
                savepath = outpath.strip("_") + "_Cyclelife"
                self.save_fig(fig, savepath)

        if not self.kwargs["cyclelife_separate_data"]:

            # Set all plot settings from Plot object
            self.fix_cyclelife(fig, ax)

            # Save fig
            savepath = outpath.strip("_") + "_Cyclelife"
            self.save_fig(fig, savepath)

    def plot_gc(self):
        """Takes all the parameters inserted in the object creation and plots Voltage-Capacity curves"""

        if self.kwargs["all_in_one"]:  # Everything goes in the same figure.

            fig, ax = self.give_fig()
            colors = [
                "tab:blue",
                "tab:orange",
                "tab:green",
                "tab:red",
                "tab:purple",
                "tab:brown",
                "tab:pink",
                "tab:gray",
                "tab:olive",
                "tab:cyan",
            ] * 5
            savepath = self.outpath

            colorbar_incrementor = -1
            for cpobj, cyc_nums, color, filename in self.file_data:
                # Get Pandas DataFrame of pot vs cap from cellpy object
                df = cpobj.get_cap(
                    method="forth-and-forth",
                    label_cycle_number=True,
                    categorical_column=True,
                )

                # Group by cycle and make list of cycle numbers
                cycgrouped = df.groupby("cycle")
                keys = []
                for key, item in cycgrouped:
                    keys.append(key)

                # Make label from filename or nickname
                if self.nicknames:
                    label = str(self.nicknames[self.files.index(filename)])
                else:
                    label = str(os.path.basename(filename))

                # Fix colorbar or cycle colors
                if self.kwargs["specific_cycles"] == None:  # Plot all cycles
                    # Set up colormap and add colorbar
                    cmap = mpl.colors.LinearSegmentedColormap.from_list(
                        "name", [color, "black"], N=256, gamma=1.0
                    )
                    norm = mpl.colors.Normalize(vmin=cyc_nums[0], vmax=cyc_nums[-1])
                    cbaxes = fig.add_axes(
                        [1.05 + colorbar_incrementor / 8, 0.1, 0.03, 0.8]
                    )
                    colorbar_incrementor += 1
                    fig.colorbar(
                        mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
                        cax=cbaxes,
                        label="Cycle number for "
                        + os.path.basename(filename).split(".")[0],
                    )
                    # fig.colorbar.ax.yaxis.get_major_locator().set_params(integer=True) #TODO fix such that we dont have decimals on the cycle colorbar!!

                # Plot cycles
                for cyc in keys:
                    if cyc in cyc_nums:
                        if self.kwargs["specific_cycles"]:
                            cyccolor = colors[0]
                            colors = colors[1:]
                        else:
                            cyccolor = cmap(cyc / keys[-1])

                        cyc_df = cycgrouped.get_group(cyc)

                        if (
                            not self.kwargs["only_dischg"]
                            and not self.kwargs["only_chg"]
                        ):
                            pass
                        elif self.kwargs["only_dischg"]:
                            dchg = cyc_df.groupby("direction")
                            cyc_df = dchg.get_group(-1)
                        elif self.kwargs["only_chg"]:
                            chg = cyc_df.groupby("direction")
                            cyc_df = chg.get_group(1)

                        # TODO: The way this is set up, when plotting both discharge and charge, the whole cycle is normalized on the maximum capacity, meaning the charge can be normalized on the discharge or the other way around.
                        if self.kwargs["galvanostatic_normalize_capacity"]:
                            # Then we normalize capacity column on the max value (since this should be max cap)
                            maxcap = cyc_df["capacity"].max()
                            cyc_df["capacity"] = cyc_df["capacity"].div(maxcap)
                            ax.set_xlabel("Normalized Capacity")

                        ax.plot(
                            cyc_df["capacity"],
                            cyc_df["voltage"],
                            label=label + ", Cyc " + str(cyc),
                            c=cyccolor,
                        )

                savepath += os.path.basename(filename).split(".")[0]

            fig.suptitle("Galvanostatic cyclingdata")
            self.fix_gc(fig, ax)

            # Save fig
            savepath += "_GC-plot"
            self.save_fig(fig, savepath)

        else:  # Then each data goes in its own figure
            for cpobj, cyc_nums, color, filename in self.file_data:
                fig, ax = self.give_fig()

                # Get Pandas DataFrame of pot vs cap from cellpy object
                df = cpobj.get_cap(
                    method="forth-and-forth",
                    label_cycle_number=True,
                    categorical_column=True,
                )

                # Group by cycle and make list of cycle numbers
                cycgrouped = df.groupby("cycle")
                keys = []
                for key, item in cycgrouped:
                    keys.append(key)

                # Fix colorbar or cycle colors
                if self.kwargs["specific_cycles"] == None:  # Plot all cycles
                    # Set up colormap and add colorbar
                    cmap = mpl.colors.LinearSegmentedColormap.from_list(
                        "name", [color, "black"], N=256, gamma=1.0
                    )
                    norm = mpl.colors.Normalize(vmin=cyc_nums[0], vmax=cyc_nums[-1])
                    fig.colorbar(
                        mpl.cm.ScalarMappable(norm=norm, cmap=cmap), label="Cycle"
                    )
                    # fig.colorbar.ax.yaxis.get_major_locator().set_params(integer=True) #TODO fix such that we dont have decimals on the cycle colorbar!!

                # Make label from filename or nickname
                if self.nicknames:
                    label = str(self.nicknames[self.files.index(filename)])
                else:
                    label = str(os.path.basename(filename))

                # Plot cycles
                colors = [
                    "tab:blue",
                    "tab:orange",
                    "tab:green",
                    "tab:red",
                    "tab:purple",
                    "tab:brown",
                    "tab:pink",
                    "tab:gray",
                    "tab:olive",
                    "tab:cyan",
                ]
                for cyc in keys:
                    if cyc in cyc_nums:
                        if self.kwargs["specific_cycles"]:
                            cyccolor = colors[0]
                            colors = colors[1:]
                        else:
                            cyccolor = cmap(cyc / keys[-1])

                        cyc_df = cycgrouped.get_group(cyc)
                        # TODO: This if elif block is pretty much the same as the one above (for all in one plot), can it be reused in stead of written twice?
                        if (
                            not self.kwargs["only_dischg"]
                            and not self.kwargs["only_chg"]
                        ):
                            pass
                        elif self.kwargs["only_dischg"]:
                            dchg = cyc_df.groupby("direction")
                            cyc_df = dchg.get_group(-1)
                        elif self.kwargs["only_chg"]:
                            chg = cyc_df.groupby("direction")
                            cyc_df = chg.get_group(1)

                        # TODO: The way this is set up, when plotting both discharge and charge, the whole cycle is normalized on the maximum capacity, meaning the charge can be normalized on the discharge or the other way around.
                        if self.kwargs["galvanostatic_normalize_capacity"]:
                            # Then we normalize capacity column on the max value (since this should be max cap)
                            maxcap = cyc_df["capacity"].max()
                            cyc_df["capacity"] = cyc_df["capacity"].div(maxcap)
                            ax.set_xlabel("Normalized Capacity")

                        ax.plot(
                            cyc_df["capacity"],
                            cyc_df["voltage"],
                            label=label.split(".")[0] + ", Cyc " + str(cyc),
                            c=cyccolor,
                        )

                # Set all plot settings from Plot object

                fig.suptitle(label)
                self.fix_gc(fig, ax)

                # Save fig
                savepath = (
                    self.outpath + os.path.basename(filename).split(".")[0] + "_GC-plot"
                )
                self.save_fig(fig, savepath)

    def plot_dQdV(self):
        """Takes all the parameters inserted in the object creation and plots dQdV"""
        from cellpy.utils import ica

        if self.kwargs["all_in_one"]:  # Everything goes in the same figure.

            fig, ax = self.give_fig()
            colors = [
                "tab:blue",
                "tab:orange",
                "tab:green",
                "tab:red",
                "tab:purple",
                "tab:brown",
                "tab:pink",
                "tab:gray",
                "tab:olive",
                "tab:cyan",
            ] * 5
            savepath = self.outpath

            colorbar_incrementor = -1
            for cpobj, cyc_nums, color, filename in self.file_data:
                # Get Pandas DataFrame of dQdV
                if self.kwargs["only_dischg"]:
                    _, df = ica.dqdv_frames(cpobj, split=True)
                elif self.kwargs["only_chg"]:
                    df, _ = ica.dqdv_frames(cpobj, split=True)
                else:
                    df = ica.dqdv_frames(cpobj)

                # Group by cycle and make list of cycle numbers
                cycgrouped = df.groupby("cycle")
                keys = []
                for key, item in cycgrouped:
                    keys.append(key)

                # Fix colorbar or cycle colors
                if self.kwargs["specific_cycles"] == None:  # Plot all cycles
                    # Set up colormap and add colorbar
                    cmap = mpl.colors.LinearSegmentedColormap.from_list(
                        "name", [color, "black"], N=256, gamma=1.0
                    )
                    norm = mpl.colors.Normalize(vmin=cyc_nums[0], vmax=cyc_nums[-1])
                    cbaxes = fig.add_axes(
                        [1.05 + colorbar_incrementor / 8, 0.1, 0.03, 0.8]
                    )
                    colorbar_incrementor += 1
                    fig.colorbar(
                        mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
                        cax=cbaxes,
                        label="Cycle number for "
                        + os.path.basename(filename).split(".")[0],
                    )

                    # fig.colorbar.ax.yaxis.get_major_locator().set_params(integer=True) #TODO fix such that we dont have decimals on the cycle colorbar!!

                # Plot cycles
                for cyc in keys:
                    if cyc in cyc_nums:
                        if self.kwargs["specific_cycles"]:
                            cyccolor = colors[0]
                            colors = colors[1:]
                        else:
                            cyccolor = cmap(cyc / keys[-1])

                        cyc_df = cycgrouped.get_group(cyc)
                        ax.plot(
                            cyc_df["voltage"],
                            cyc_df["dq"],
                            label=os.path.basename(filename).split(".")[0]
                            + ", Cyc "
                            + str(cyc),
                            c=cyccolor,
                        )
                savepath += os.path.basename(filename).split(".")[0]

            fig.suptitle("dQdV")
            self.fix_dqdv(fig, ax)

            # Save fig
            savepath += "_dQdV-plot"
            self.save_fig(fig, savepath)

        else:
            for cpobj, cyc_nums, color, filename in self.file_data:
                fig, ax = self.give_fig()

                # Get Pandas DataFrame of dQdV
                if self.kwargs["only_dischg"]:
                    _, df = ica.dqdv_frames(cpobj, split=True)
                elif self.kwargs["only_chg"]:
                    df, _ = ica.dqdv_frames(cpobj, split=True)
                else:
                    df = ica.dqdv_frames(cpobj)

                # Group by cycle and make list of cycle numbers
                cycgrouped = df.groupby("cycle")
                keys = []
                for key, item in cycgrouped:
                    keys.append(key)

                # Create the plot obj
                fig, ax = plt.subplots(figsize=(6, 4))

                # Fix colorbar or cycle colors
                if self.kwargs["specific_cycles"] == None:  # Plot all cycles
                    # Set up colormap and add colorbar
                    cmap = mpl.colors.LinearSegmentedColormap.from_list(
                        "name", [color, "black"], N=256, gamma=1.0
                    )
                    norm = mpl.colors.Normalize(vmin=cyc_nums[0], vmax=cyc_nums[-1])
                    fig.colorbar(
                        mpl.cm.ScalarMappable(norm=norm, cmap=cmap), label="Cycle"
                    )
                    # fig.colorbar.ax.yaxis.get_major_locator().set_params(integer=True) #TODO fix such that we dont have decimals on the cycle colorbar!!

                # Plot cycles
                colors = [
                    "tab:blue",
                    "tab:orange",
                    "tab:green",
                    "tab:red",
                    "tab:purple",
                    "tab:brown",
                    "tab:pink",
                    "tab:gray",
                    "tab:olive",
                    "tab:cyan",
                ]
                for cyc in keys:
                    if cyc in cyc_nums:
                        if self.kwargs["specific_cycles"]:
                            cyccolor = colors[0]
                            colors = colors[1:]
                        else:
                            cyccolor = cmap(cyc / keys[-1])
                        cyc_df = cycgrouped.get_group(cyc)
                        ax.plot(
                            cyc_df["voltage"],
                            cyc_df["dq"],
                            label="Cycle " + str(cyc),
                            c=cyccolor,
                        )

                # Set all plot settings from Plot object
                fig.suptitle(os.path.basename(filename))
                self.fix_dqdv(fig, ax)

                # Save fig
                savepath = (
                    self.outpath
                    + os.path.basename(filename).split(".")[0]
                    + "_dQdV-plot"
                )
                self.save_fig(fig, savepath)

    def plot_gc_and_dQdV(self):
        """Takes all the parameters inserted in the object creation and plots Voltage-Curves and dQdV data together"""
        from cellpy.utils import ica

        if self.kwargs["all_in_one"]:  # Everything goes in the same figure.
            fig, ax = self.give_fig()
            fig.delaxes(ax)
            ax1, ax2 = fig.subplots(1, 2, sharey=True)
            fig.set_size_inches(8, 4)
            fig.subplots_adjust(wspace=0)

            colors = [
                "tab:blue",
                "tab:orange",
                "tab:green",
                "tab:red",
                "tab:purple",
                "tab:brown",
                "tab:pink",
                "tab:gray",
                "tab:olive",
                "tab:cyan",
            ] * 5
            savepath = self.outpath

            colorbar_incrementor = -1
            for cpobj, cyc_nums, color, filename in self.file_data:

                # Get Pandas DataFrame of pot vs cap from cellpy object
                df = cpobj.get_cap(
                    method="forth-and-forth",
                    label_cycle_number=True,
                    categorical_column=True,
                )

                # Group by cycle and make list of cycle numbers
                cycgrouped = df.groupby("cycle")
                keys = []
                for key, item in cycgrouped:
                    keys.append(key)

                # Fix colorbar or cycle colors
                if self.kwargs["specific_cycles"] == None:  # Plot all cycles
                    # Set up colormap and add colorbar
                    cmap = mpl.colors.LinearSegmentedColormap.from_list(
                        "name", [color, "black"], N=256, gamma=1.0
                    )
                    norm = mpl.colors.Normalize(vmin=cyc_nums[0], vmax=cyc_nums[-1])
                    cbaxes = fig.add_axes(
                        [1.05 + colorbar_incrementor / 8, 0.1, 0.03, 0.8]
                    )
                    colorbar_incrementor += 1
                    fig.colorbar(
                        mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
                        cax=cbaxes,
                        label="Cycle number for "
                        + os.path.basename(filename).split(".")[0],
                        pad=0.2,
                    )

                # Plot GC in leftmost plot (ax)
                for cyc in keys:
                    if cyc in cyc_nums:
                        if self.kwargs["specific_cycles"]:
                            cyccolor = colors[0]
                            colors = colors[1:]
                        else:
                            cyccolor = cmap(cyc / keys[-1])

                        cyc_df = cycgrouped.get_group(cyc)
                        if (
                            not self.kwargs["only_dischg"]
                            and not self.kwargs["only_chg"]
                        ):
                            ax1.plot(
                                cyc_df["capacity"],
                                cyc_df["voltage"],
                                label=os.path.basename(filename).split(".")[0]
                                + ", Cyc "
                                + str(cyc),
                                c=cyccolor,
                            )
                        elif self.kwargs["only_dischg"]:
                            dchg = cyc_df.groupby("direction")
                            dchg_df = dchg.get_group(-1)
                            ax1.plot(
                                dchg_df["capacity"],
                                dchg_df["voltage"],
                                label=os.path.basename(filename).split(".")[0]
                                + ", Cyc "
                                + str(cyc),
                                c=cyccolor,
                            )
                        elif self.kwargs["only_chg"]:
                            chg = cyc_df.groupby("direction")
                            chg_df = chg.get_group(1)
                            ax1.plot(
                                chg_df["capacity"],
                                chg_df["voltage"],
                                label=os.path.basename(filename).split(".")[0]
                                + ", Cyc "
                                + str(cyc),
                                c=cyccolor,
                            )

                # Get Pandas DataFrame for dQdV
                if self.kwargs["only_dischg"]:
                    _, df = ica.dqdv_frames(cpobj, split=True)
                elif self.kwargs["only_chg"]:
                    df, _ = ica.dqdv_frames(cpobj, split=True)
                else:
                    df = ica.dqdv_frames(cpobj)

                # Group by cycle and make list of cycle numbers
                cycgrouped = df.groupby("cycle")
                keys = []
                for key, item in cycgrouped:
                    keys.append(key)

                # Plot cycles
                for cyc in keys:
                    if cyc in cyc_nums:
                        if self.kwargs["specific_cycles"]:
                            cyccolor = colors[0]
                            colors = colors[1:]
                        else:
                            cyccolor = cmap(cyc / keys[-1])

                        cyc_df = cycgrouped.get_group(cyc)
                        ax2.plot(
                            cyc_df["dq"],
                            cyc_df["voltage"],
                            label=os.path.basename(filename).split(".")[0]
                            + ", Cyc "
                            + str(cyc),
                            c=cyccolor,
                        )
                savepath += os.path.basename(filename).split(".")[0]

            # Set all plot settings from Plot object
            fig.suptitle("GC and dQdV")
            self.fix_gc_and_dqdv(fig, [ax1, ax2])

            # Save fig
            savepath = savepath + "_GC-dQdV-plot"
            self.save_fig(fig, savepath)

        else:  # Then all files are placed in separate plots
            for cpobj, cyc_nums, color, filename in self.file_data:
                fig, ax = self.give_fig()
                fig.delaxes(ax)
                ax1, ax2 = fig.subplots(1, 2, sharey=True)
                fig.set_size_inches(8, 4)
                fig.subplots_adjust(wspace=0)

                colors = [
                    "tab:blue",
                    "tab:orange",
                    "tab:green",
                    "tab:red",
                    "tab:purple",
                    "tab:brown",
                    "tab:pink",
                    "tab:gray",
                    "tab:olive",
                    "tab:cyan",
                ] * 5

                # Get Pandas DataFrame of pot vs cap from cellpy object
                df = cpobj.get_cap(
                    method="forth-and-forth",
                    label_cycle_number=True,
                    categorical_column=True,
                )

                # Group by cycle and make list of cycle numbers
                cycgrouped = df.groupby("cycle")
                keys = []
                for key, item in cycgrouped:
                    keys.append(key)

                # Fix colorbar or cycle colors
                if self.kwargs["specific_cycles"] == None:  # Plot all cycles
                    # Set up colormap and add colorbar
                    cmap = mpl.colors.LinearSegmentedColormap.from_list(
                        "name", [color, "black"], N=256, gamma=1.0
                    )
                    norm = mpl.colors.Normalize(vmin=cyc_nums[0], vmax=cyc_nums[-1])
                    fig.colorbar(
                        mpl.cm.ScalarMappable(norm=norm, cmap=cmap), label="Cycle"
                    )

                # Plot GC in leftmost plot (ax)
                for cyc in keys:
                    if cyc in cyc_nums:
                        if self.kwargs["specific_cycles"]:
                            cyccolor = colors[0]
                            colors = colors[1:]
                        else:
                            cyccolor = cmap(cyc / keys[-1])

                        cyc_df = cycgrouped.get_group(cyc)
                        if (
                            not self.kwargs["only_dischg"]
                            and not self.kwargs["only_chg"]
                        ):
                            ax1.plot(
                                cyc_df["capacity"],
                                cyc_df["voltage"],
                                label=os.path.basename(filename).split(".")[0]
                                + ", Cyc "
                                + str(cyc),
                                c=cyccolor,
                            )
                        elif self.kwargs["only_dischg"]:
                            dchg = cyc_df.groupby("direction")
                            dchg_df = dchg.get_group(-1)
                            ax1.plot(
                                dchg_df["capacity"],
                                dchg_df["voltage"],
                                label=os.path.basename(filename).split(".")[0]
                                + ", Cyc "
                                + str(cyc),
                                c=cyccolor,
                            )
                        elif self.kwargs["only_chg"]:
                            chg = cyc_df.groupby("direction")
                            chg_df = chg.get_group(1)
                            ax1.plot(
                                chg_df["capacity"],
                                chg_df["voltage"],
                                label=os.path.basename(filename).split(".")[0]
                                + ", Cyc "
                                + str(cyc),
                                c=cyccolor,
                            )

                # Get Pandas DataFrame for dQdV
                if self.kwargs["only_dischg"]:
                    _, df = ica.dqdv_frames(cpobj, split=True)
                elif self.kwargs["only_chg"]:
                    df, _ = ica.dqdv_frames(cpobj, split=True)
                else:
                    df = ica.dqdv_frames(cpobj)

                # Group by cycle and make list of cycle numbers
                cycgrouped = df.groupby("cycle")
                keys = []
                for key, item in cycgrouped:
                    keys.append(key)

                # Plot cycles
                for cyc in keys:
                    if cyc in cyc_nums:
                        if self.kwargs["specific_cycles"]:
                            cyccolor = colors[0]
                            colors = colors[1:]
                        else:
                            cyccolor = cmap(cyc / keys[-1])

                        cyc_df = cycgrouped.get_group(cyc)
                        ax2.plot(
                            cyc_df["dq"],
                            cyc_df["voltage"],
                            label=os.path.basename(filename).split(".")[0]
                            + ", Cyc "
                            + str(cyc),
                            c=cyccolor,
                        )

                # Set all plot settings from Plot object
                fig.suptitle(os.path.basename(filename))
                self.fix_gc_and_dqdv(fig, [ax1, ax2])

                # Save fig
                savepath = (
                    self.outpath
                    + os.path.basename(filename).split(".")[0]
                    + "_GC-dQdV-plot"
                )
                self.save_fig(fig, savepath)

        """# Fix colorbar or cycle colors
        if not specific_cycles: # If this is none, then plot all!
            # Set up colormap and add colorbar
            cmap = mpl.colors.LinearSegmentedColormap.from_list("name", [color, "black"], N=256, gamma=1.0)
            norm = mpl.colors.Normalize(vmin=cyc_nums[0], vmax=cyc_nums[-1])
            fig.colorbar(mpl.cm.ScalarMappable(norm=norm, cmap=cmap),label='Cycle')

        ## Plot GC on the left subplot (ax[0]) ##

        # Get Pandas DataFrame of pot vs cap from cellpy object
        df = cpobj.get_cap(method="forth-and-forth", label_cycle_number=True, categorical_column=True)

        # Group by cycle and make list of cycle numbers
        cycgrouped = df.groupby("cycle")
        keys = []
        for key, item in cycgrouped:
            keys.append(key)

        # Plot cycles
        colors =  ['tab:blue', 'tab:orange', 'tab:green', 'tab:red', 'tab:purple', 'tab:brown', 'tab:pink', 'tab:gray', 'tab:olive', 'tab:cyan' ]
        for cyc in keys:
            if cyc in cyc_nums:
                if specific_cycles:
                    cyccolor = colors[0]
                    colors = colors[1:]
                else:
                    cyccolor = cmap(cyc/keys[-1])
                cyc_df = cycgrouped.get_group(cyc)
                axs[0].plot(cyc_df["capacity"], cyc_df["voltage"], label="Cycle " + str(cyc), c = cyccolor)


        ## Plot dQdV on the right subplot (ax[1]) ##

        from cellpy.utils import ica
        # Get Pandas DataFrame of pot vs cap from cellpy object
        df = ica.dqdv_frames(cpobj)

        # Group by cycle and make list of cycle numbers
        cycgrouped = df.groupby("cycle")
        keys = []
        for key, item in cycgrouped:
            keys.append(key)

        # Plot cycles
        colors =  ['tab:blue', 'tab:orange', 'tab:green', 'tab:red', 'tab:purple', 'tab:brown', 'tab:pink', 'tab:gray', 'tab:olive', 'tab:cyan' ]
        for cyc in keys:
            if cyc in cyc_nums:
                if specific_cycles:
                    cyccolor = colors[0]
                    colors = colors[1:]
                else:
                    cyccolor = cmap(cyc/keys[-1])
                cyc_df = cycgrouped.get_group(cyc)
                axs[1].plot(cyc_df["dq"], cyc_df["voltage"], label=str(cyc), c = cyccolor)

        # Set all plot settings from Plot object
        fig.suptitle(os.path.basename(file))
        self.fix_gc_and_dqdv(fig, axs)

        # Save fig
        savepath = self.outpath + os.path.basename(file).split(".")[0] + "_GC-dQdV-plot"
        print("Saving to: " + savepath)
        fig.savefig(savepath, bbox_inches='tight')"""

    def plot_cap_from_rc(self):
        """Takes all the parameters inserted in the object creation and plots capacity VS inverse c-rate"""
        # Spawn fig and axis for plotting
        fig, ax = self.give_fig()

        # Get labels and handles for legend generation and eventual savefile
        handles, labels = ax.get_legend_handles_labels()
        # handles.append(Line2D([0], [0], marker='o', color='black', alpha = 0.2, label = 'Charge capacity', linestyle=''))
        # handles.append(Line2D([0], [0], marker='o', color='black', alpha = 0.2, label = 'Disharge capacity', linestyle=''))
        # handles.append(Line2D([0], [0], marker='+', color='black', label = 'Cap avg per C-rate', linestyle=''))

        outpath = self.outpath
        for cpobj, cyc_nums, color, filename in self.file_data:
            # Get Pandas DataFrame of pot vs cap from cellpy object
            # df = cpobj.get_cap(method="forth-and-forth", label_cycle_number=True, categorical_column=True)
            outpath += os.path.basename(filename).split(".")[0] + "_"

            handles.append(
                Line2D([0], [0], marker="o", color=color, label=filename, linestyle="")
            )

            stepstable = cpobj.cell.steps
            chglist, dchglist = get_effective_C_rates_and_caps(stepstable)

            # Remove all cycles which are not in cyc_nums by looking at the 0th element (cyc num) of every sublist in chglist
            new_chglist = [x for x in chglist if x[0] in cyc_nums]
            new_dchglist = [x for x in dchglist if x[0] in cyc_nums]

            linregress_xlist = []
            linregress_ylist = []
            for chg, dchg in zip(new_chglist, new_dchglist):
                # print(dchg)
                # ax.scatter(chg[1] , chg[2] , color = color, alpha = 0.2)
                ax.scatter(1 / dchg[1], dchg[2], color=color, alpha=1)

                linregress_xlist.append(1 / dchg[1])
                linregress_ylist.append(dchg[2])

            # print(linregress_ylist)
            # Fitting curve to the exponential function
            # Import curve fitting package from scipy
            # from scipy.optimize import curve_fit

            x_arr = np.array(linregress_xlist)
            y_arr = np.array(linregress_ylist)

            # Average the capacity for each c-rate
            def _reduce_to_averages(xvals, yvals):
                """This function scans through the data and averages relevant points together."""

                point_grouped = []
                point_lst = []
                dists = []
                for i in range(1, len(xvals)):
                    prev_point = np.array((xvals[i - 1], yvals[i - 1]))
                    curr_point = np.array((xvals[i], yvals[i]))

                    dev = 0.3
                    if (
                        (prev_point * (1 - dev))[0]
                        < curr_point[0]
                        < (prev_point * (1 + dev))[0]
                    ):
                        # If this point is within dev (percentage sort of) of last point, then its in the same c-rate
                        point_lst.append(curr_point)
                    else:
                        # New c-rate
                        point_grouped.append(point_lst)
                        point_lst = []

                print(point_grouped)

                x_arr = []
                y_arr = []

                for group in point_grouped:
                    stacked_arr = np.stack(group, axis=1)
                    averaged_arr = np.average(stacked_arr, axis=1)
                    x_arr.append(averaged_arr[0])
                    y_arr.append(averaged_arr[1])

                print(x_arr)
                print(y_arr)

                return x_arr, y_arr

            # x_arr, y_arr = _reduce_to_averages(x_arr, y_arr)

            # ax.scatter(x_arr, y_arr, marker="+")
            # def _exp_func(x,a,b,c):
            #    return -a* (b**x) + a + -a * (b**(x+c)) +a

            # pars, cov = curve_fit(f=_exp_func, p0 = [50, 0.7, 0], xdata = x_arr, ydata=y_arr, bounds = ([0,0.1, -20],[1e9, 1, 20]))

            # x_vals = np.linspace(min(x_arr), max(x_arr), 100) #x_arr[0], x_arr[-1], 100)
            # ax.plot(x_vals, _exp_func(x_vals, *pars))
            # ax.hlines(max(y_arr), ax.get_xlim()[0], ax.get_xlim()[1], colors = color, linestyle='--')
            # Get the standard deviations of the parameters (square roots of the # diagonal of the covariance)
            # std_dev = np.sqrt(np.diag(cov))
            # Make a sweet legend to put on this
            # handles.append(
            #    Line2D(
            #        [0], [0],
            #        marker="_", color=color,
            #        label = 'Calculated maximum capacity:' + '\n' +'{:.2e} $\pm$ {:.2e}'.format(pars[0], std_dev[0]) + r'$\left[\mu Ah\right]$', linestyle=''
            #        ))

            ax.hlines(
                max(y_arr),
                ax.get_xlim()[0],
                ax.get_xlim()[1],
                colors=color,
                linestyle="--",
            )
            handles.append(
                Line2D(
                    [0],
                    [0],
                    marker="_",
                    color=color,
                    label="Highest capacity:"
                    + "\n"
                    + "{:.2e}".format(max(y_arr))
                    + r"$\left[\mu Ah\right]$",
                    linestyle="",
                )
            )

        self.fix_cap_from_rc(fig, ax, handles)

        # Save fig
        savepath = outpath + "CapDet"
        self.save_fig(fig, savepath)

    def fix_cyclelife(self, fig, ax):
        """Makes the finishing touches to the cyclelife plot"""
        # Applies kwargs settings and other plot settings

        ## Parameters which could be user defined later
        """
        ax.set(
            xticks = (np.arange(0, 150), step=20)),
            yticks = (np.arange(3, 5, step=0.2)),
            )
        """
        # Get labels and handles for legend generation and eventual savefile
        handles, labels = ax.get_legend_handles_labels()
        if not self.kwargs["only_dischg"]:
            handles.append(
                Line2D(
                    [0],
                    [0],
                    marker="o",
                    color="black",
                    alpha=0.2,
                    label="Charge capacity",
                    linestyle="",
                )
            )

        if self.kwargs["cyclelife_coulombic_efficiency"]:
            handles.append(
                Line2D(
                    [0],
                    [0],
                    marker="+",
                    color="black",
                    alpha=1,
                    label="Coulombic Efficiency",
                    linestyle="",
                )
            )

        if (
            self.kwargs["cyclelife_charge_c_rate"]
            and not self.kwargs["cyclelife_discharge_c_rate"]
        ):
            handles.append(
                Line2D(
                    [0],
                    [0],
                    marker="_",
                    color="black",
                    alpha=1,
                    label="Effective charge C-rate",
                    linestyle="",
                )
            )
        elif (
            not self.kwargs["cyclelife_charge_c_rate"]
            and self.kwargs["cyclelife_discharge_c_rate"]
        ):
            handles.append(
                Line2D(
                    [0],
                    [0],
                    marker="_",
                    color="black",
                    alpha=1,
                    label="Effective discharge C-rate",
                    linestyle="",
                )
            )
        elif (
            self.kwargs["cyclelife_charge_c_rate"]
            and self.kwargs["cyclelife_discharge_c_rate"]
        ):
            handles.append(
                Line2D(
                    [0],
                    [0],
                    marker="_",
                    color="black",
                    alpha=1,
                    label="Effective charge C-rate",
                    linestyle="",
                )
            )
            handles.append(
                Line2D(
                    [0],
                    [0],
                    marker="_",
                    color="black",
                    alpha=0.2,
                    label="Effective discharge C-rate",
                    linestyle="",
                )
            )

        # The params below should always be like this.
        ax.tick_params(direction="in", top="true", right="true")
        ax.xaxis.get_major_locator().set_params(integer=True)

        # Apply all kwargs to plot
        try:
            # Cyclelife plot details
            ax.set(xlabel=self.kwargs["cyclelife_xlabel"])
            if self.kwargs["cyclelife_percentage"]:
                ax.set(ylabel=self.kwargs["cyclelife_ylabel_percent"])
            else:
                ax.set(ylabel=self.kwargs["cyclelife_ylabel"])

            # General plot details
            fig.set_size_inches(self.kwargs["figsize"])
            if type(self.kwargs["figtitle"]) == str:
                fig.suptitle(self.kwargs["figtitle"])
            else:
                fig.suptitle("Capacity versus Cycle life")

        except Exception as e:
            logging.error(e)

        # Take care of having the legend outside the plot
        if self.kwargs["cyclelife_legend_outside"]:
            if (
                self.kwargs["cyclelife_coulombic_efficiency"]
                or self.kwargs["cyclelife_charge_c_rate"]
                or self.kwargs["cyclelife_discharge_c_rate"]
            ):
                ax.legend(handles=handles, bbox_to_anchor=(1.18, 1), loc="upper left")
            else:
                ax.legend(handles=handles, bbox_to_anchor=(1.05, 1), loc="upper left")

            figsize = self.kwargs["figsize"]
            fig.set_size_inches((figsize[0] + 3, figsize[1]))
        else:
            ax.legend(handles=handles)
        fig.tight_layout()  # Needed to not clip ylabel on coulombic efficiency

    def fix_cap_from_rc(self, fig, ax, handles):
        """Makes the finishing touches to the capacity vs inverse C-rate plot"""
        ax.tick_params(direction="in", top="true", right="true")
        ax.set(
            xlabel=r"Inverse C-rate $\left[ h \right]$",
            ylabel=r"Capacity $\left[\mu Ah \right]$",
        )
        # General plot details
        fig.set_size_inches(self.kwargs["figsize"])
        if type(self.kwargs["figtitle"]) == str:
            fig.suptitle(self.kwargs["figtitle"])
        else:
            fig.suptitle("Capacity determination from Rate Capability")

        # Take care of having the legend outside the plot
        if self.kwargs["cyclelife_legend_outside"]:
            ax.legend(handles=handles, bbox_to_anchor=(1.05, 1), loc="upper left")

            figsize = self.kwargs["figsize"]
            fig.set_size_inches((figsize[0] + 3, figsize[1]))
        else:
            ax.legend(handles=handles)
        fig.tight_layout()  # Needed to not clip ylabel on coulombic efficiency

    def fix_gc(self, fig, ax):
        """Makes the finishing touches to the voltage-curves plot"""
        # Applies kwargs settings and other plot settings

        ## Parameters which could be user defined later
        """
        ax.set(
            xticks = (np.arange(0, 150), step=20)),
            yticks = (np.arange(3, 5, step=0.2)),
            )
        """

        # The params below should always be like this.
        ax.tick_params(direction="in", top="true", right="true")

        # Apply all kwargs to plot
        try:
            # Galvanostatic plot details
            ax.set(xlabel=self.kwargs["galvanostatic_xlabel"])
            ax.set(ylabel=self.kwargs["galvanostatic_ylabel"])
            ax.set(ylim=self.kwargs["galvanostatic_potlim"])
            ax.set(xlim=self.kwargs["galvanostatic_caplim"])

            if self.kwargs["specific_cycles"] != None:
                ax.legend()

            # General plot details
            fig.set_size_inches(self.kwargs["figsize"])
            if type(self.kwargs["figtitle"]) == str:
                fig.suptitle(self.kwargs["figtitle"])

        except Exception as e:
            logging.error(e)

    def fix_dqdv(self, fig, ax):
        """Makes the finishing touches to the dQdV plot"""
        # Applies kwargs settings and other plot settings

        ## Parameters which could be user defined later
        """
        ax.set(
            xticks = (np.arange(0, 150), step=20)),
            yticks = (np.arange(3, 5, step=0.2)),
            )
        """

        # The params below should always be like this.
        ax.tick_params(direction="in", top="true", right="true")

        # Apply all kwargs to plot
        try:
            # Cyclelife plot details
            ax.set(xlabel=self.kwargs["dqdv_xlabel"])
            ax.set(ylabel=self.kwargs["dqdv_ylabel"])
            ax.set(ylim=self.kwargs["dqdv_dqlim"])
            ax.set(xlim=self.kwargs["dqdv_potlim"])

            if self.kwargs["specific_cycles"] != None:
                ax.legend()

            # General plot details
            fig.set_size_inches(self.kwargs["figsize"])
            if type(self.kwargs["figtitle"]) == str:
                fig.suptitle(self.kwargs["figtitle"])

        except Exception as e:
            logging.error(e)

    def fix_gc_and_dqdv(self, fig, axs):
        """Makes the finishing touches to the dQdV / Voltage curves plot"""
        for ax in axs:
            # The params below should always be like this.
            ax.tick_params(direction="in", top="true", right="true")

        # Apply all kwargs to plot
        try:
            # dQdV plot details
            axs[1].set(
                xlabel=self.kwargs["dqdv_ylabel"]
            )  # switched x and y label since this dQdV plot is flipped to match the adjacent gc plot
            axs[1].set(ylabel="")  # Empty since we already have potential on gc axs
            axs[1].set(ylim=self.kwargs["galvanostatic_potlim"])
            axs[1].set(xlim=self.kwargs["dqdv_dqlim"])

            # Galvanostatic plot details
            axs[0].set(xlabel=self.kwargs["galvanostatic_xlabel"])
            axs[0].set(ylabel=self.kwargs["galvanostatic_ylabel"])
            axs[0].set(ylim=self.kwargs["galvanostatic_potlim"])
            axs[0].set(xlim=self.kwargs["galvanostatic_caplim"])

            if self.kwargs["specific_cycles"] != None:
                axs[0].legend()

            # General plot details
            fig.set_size_inches(self.kwargs["figsize"])
            if type(self.kwargs["figtitle"]) == str:
                fig.suptitle(self.kwargs["figtitle"])

        except Exception as e:
            print(e)
            logging.error(e)

    def save_fig(self, fig, savepath):
        """The point of this is to have savefig parameters the same across
        all plots (for now just fig dpi and bbox inches)"""
        if self.kwargs.get("save_figures", True):
            if self.kwargs["outname"]:
                savepath = (
                    self.kwargs["outpath"]
                    + self.kwargs["outname"]
                    + self.kwargs["outtype"]
                )
            else:
                savepath += self.kwargs["outtype"]

            print("Saving to: " + savepath)
            fig.savefig(savepath, bbox_inches="tight", dpi=self.kwargs["figres"])


def get_effective_C_rates(steptable):

    newdf = steptable[["step_time_avr", "cycle", "type"]]
    chg_c_rates = []
    dchg_c_rates = []
    for i, elem in enumerate(newdf.iterrows()):
        if elem[1]["type"] == "charge":
            chg_c_rates.append(1 / (elem[1]["step_time_avr"] / 3600))
        elif elem[1]["type"] == "discharge":
            dchg_c_rates.append(1 / (elem[1]["step_time_avr"] / 3600))

    return chg_c_rates, dchg_c_rates


def get_effective_C_rates_and_caps(steptable):
    newdf = steptable[
        ["step_time_avr", "cycle", "type", "charge_avr", "discharge_last"]
    ]
    chglist = (
        []
    )  # [[cycle, chg_crate, chg_cap], [cycle increase with crates and capacities for this cycle]]
    dchglist = []
    for i, elem in enumerate(newdf.iterrows()):
        cyc = elem[1]["cycle"]

        if elem[1]["type"] == "charge":
            chglist.append(
                [
                    cyc,
                    1 / (elem[1]["step_time_avr"] / 3600),
                    elem[1]["charge_avr"] * 1000,
                ]
            )
        elif elem[1]["type"] == "discharge":
            dchglist.append(
                [
                    cyc,
                    1 / (elem[1]["step_time_avr"] / 3600),
                    elem[1]["discharge_last"] * 1000 * 1000,
                ]
            )

    return chglist, dchglist


def main():
    log.setup_logging(default_level="DEBUG")
    f1 = Path("../../testdata/data/20160805_test001_45_cc_01.res")
    f2 = Path("../../testdata/data/20160805_test001_47_cc_01.res")

    raw_files = [f1, f2]
    nicknames = ["cell1", "cell2"]

    logging.debug(raw_files)
    logging.debug(nicknames)

    ezplt = EasyPlot(raw_files, nicknames, figtitle="Test1", save_figures=True)
    ezplt.plot()
    plt.show()

    return


def _dev_journal_loading():
    log.setup_logging(default_level="DEBUG")
    journal_file = Path("../../testdata/db/cellpy_batch_test.json")
    ezplt = EasyPlot(
        None,
        journal=journal_file,
        figtitle="Test1",
        save_figures=False,
        save_journal=True,
        outpath="./tmp/",
    )
    ezplt.plot()
    # plt.show()

    return


if __name__ == "__main__":
    print(" running easyplot ".center(80, "-"))
    _dev_journal_loading()
    print(" finished ".center(80, "-"))



""" Very simple implementation of a plugin-like infrastructure"""
from dataclasses import dataclass, field
from importlib import import_module
from pathlib import Path

from ruamel import yaml

# TODO: make tests.
# TODO: move this into its own module (not __init__).
# TODO: document for devs.
# TODO: Bonus - make a python package/pre-commit hook that
# turns TODO-statements into issues.

HARD_CODED_MODULE_PATH = "cellpy.readers.instruments.configurations"
OPTIONAL_DICTIONARY_ATTRIBUTE_NAMES = [
    "file_info",
    "formatters",
    "prefixes",
    "pre_processors",
    "post_processors",
    "meta_keys",
    "incremental_unit_labels",
    "not_implemented_in_cellpy_yet_renaming_dict",
    "raw_units",
    "raw_limits",
    "states",
    "unit_labels",
]
OPTIONAL_LIST_ATTRIBUTE_NAMES = [
    "columns_to_keep",
]


RAW_LIMITS = {
    "current_hard": 0.000_000_000_000_1,
    "current_soft": 0.000_01,
    "stable_current_hard": 2.0,
    "stable_current_soft": 4.0,
    "stable_voltage_hard": 2.0,
    "stable_voltage_soft": 4.0,
    "stable_charge_hard": 0.001,
    "stable_charge_soft": 5.0,
    "ir_change": 0.00001,
}


@dataclass
class ModelParameters:
    """Dataclass to store sub-model specific parameters."""

    name: str
    # The unit labels are only used when generating new headers
    # (normal_headers_renaming_dict)
    # with units (post_processor: get_column_names):
    file_info: dict = field(default_factory=dict)
    unit_labels: dict = field(default_factory=dict)
    incremental_unit_labels: dict = field(default_factory=dict)

    normal_headers_renaming_dict: dict = field(default_factory=dict)
    not_implemented_in_cellpy_yet_renaming_dict: dict = field(default_factory=dict)
    columns_to_keep: list = field(default_factory=list)
    states: dict = field(default_factory=dict)
    raw_units: dict = field(default_factory=dict)
    raw_limits: dict = field(default_factory=dict)
    formatters: dict = field(default_factory=dict)
    meta_keys: dict = field(default_factory=dict)
    pre_processors: dict = field(default_factory=dict)
    post_processors: dict = field(default_factory=dict)
    # used for defining e.g. M, k, etc. - probably not needed:
    prefixes: dict = field(default_factory=dict)


def register_local_configuration_from_yaml_file(instrument) -> ModelParameters:
    """register a module (.yml file) and return it.

    This function will dynamically import the given module from the
    cellpy.readers.instruments.configurations module and return it.

    Returns: ModelParameters

    """

    yml = yaml.YAML()
    with open(instrument, "r") as ff:
        settings = yml.load(ff.read())

    name = Path(instrument).name

    optional_dictionary_attributes = {
        key: settings.get(key, dict()) for key in OPTIONAL_DICTIONARY_ATTRIBUTE_NAMES
    }

    optional_list_attributes = {
        key: settings.get(key, list()) for key in OPTIONAL_LIST_ATTRIBUTE_NAMES
    }

    # special hacks
    # -- raw limits (this should be moved to internal settings, prms or something like that
    raw_limits = optional_dictionary_attributes["raw_limits"]
    if not raw_limits:
        raw_limits = RAW_LIMITS

    model_01 = ModelParameters(
        name=name,
        file_info=optional_dictionary_attributes["file_info"],
        normal_headers_renaming_dict=settings["normal_headers_renaming_dict"],
        unit_labels=optional_dictionary_attributes["unit_labels"],
        prefixes=optional_dictionary_attributes["prefixes"],
        incremental_unit_labels=optional_dictionary_attributes[
            "incremental_unit_labels"
        ],
        not_implemented_in_cellpy_yet_renaming_dict=optional_dictionary_attributes[
            "not_implemented_in_cellpy_yet_renaming_dict"
        ],
        columns_to_keep=optional_list_attributes["columns_to_keep"],
        states=optional_dictionary_attributes["states"],
        raw_units=optional_dictionary_attributes["raw_units"],
        raw_limits=raw_limits,
        meta_keys=optional_dictionary_attributes["meta_keys"],
        formatters=optional_dictionary_attributes["formatters"],
        pre_processors=optional_dictionary_attributes["pre_processors"],
        post_processors=optional_dictionary_attributes["post_processors"],
    )
    return model_01


def register_configuration_from_module(
    name: str = "one",
    module: str = "maccor_txt_one",
    _module_path=None,
    _m=None,
) -> ModelParameters:
    """register a python module (.py file) and return it.

    This function will dynamically import the given module from the
    cellpy.readers.instruments.configurations module and return it.

    Returns: ModelParameters
    """

    if _m is None:
        if _module_path is None:
            _m = import_module(f"{HARD_CODED_MODULE_PATH}.{module}")
        else:
            _m = import_module(f"{_module_path}.{module}")

    optional_dictionary_attributes = {
        key: getattr(_m, key, dict()) for key in OPTIONAL_DICTIONARY_ATTRIBUTE_NAMES
    }

    optional_list_attributes = {
        key: getattr(_m, key, list()) for key in OPTIONAL_LIST_ATTRIBUTE_NAMES
    }

    # special hacks
    # -- raw limits (this should be moved to internal settings, prms or something like that
    raw_limits = optional_dictionary_attributes["raw_limits"]
    if not raw_limits:
        raw_limits = RAW_LIMITS

    return ModelParameters(
        name=name,
        file_info=optional_dictionary_attributes["file_info"],
        normal_headers_renaming_dict=_m.normal_headers_renaming_dict,
        unit_labels=optional_dictionary_attributes["unit_labels"],
        prefixes=optional_dictionary_attributes["prefixes"],
        incremental_unit_labels=optional_dictionary_attributes[
            "incremental_unit_labels"
        ],
        not_implemented_in_cellpy_yet_renaming_dict=optional_dictionary_attributes[
            "not_implemented_in_cellpy_yet_renaming_dict"
        ],
        columns_to_keep=optional_list_attributes["columns_to_keep"],
        states=optional_dictionary_attributes["states"],
        raw_units=optional_dictionary_attributes["raw_units"],
        raw_limits=raw_limits,
        meta_keys=optional_dictionary_attributes["meta_keys"],
        formatters=optional_dictionary_attributes["formatters"],
        pre_processors=optional_dictionary_attributes["pre_processors"],
        post_processors=optional_dictionary_attributes["post_processors"],
    )



import collections
import logging
import os
import pathlib
import warnings
from copy import deepcopy

import numpy as np
import pandas as pd
from scipy import stats

import cellpy
from cellpy import prmreader, prms
from cellpy.parameters.internal_settings import (
    ATTRS_CELLPYDATA,
    ATTRS_DATASET,
    headers_journal,
    headers_normal,
    headers_step_table,
    headers_summary,
)
from cellpy.readers.cellreader import CellpyData
from cellpy.utils import batch, ica

hdr_summary = headers_summary
hdr_steps = headers_step_table
hdr_normal = headers_normal
hdr_journal = headers_journal


def _make_average(
    frames, keys=None, columns=None, normalize_cycles=False, key_index_bounds=None
):
    if key_index_bounds is None:
        key_index_bounds = [1, -2]
    hdr_norm_cycle = hdr_summary["normalized_cycle_index"]
    hdr_cum_charge = hdr_summary["cumulated_charge_capacity"]
    cell_id = ""
    not_a_number = np.NaN
    new_frames = []
    if columns is None:
        columns = frames[0].columns

    if keys is not None:
        if isinstance(keys, (list, tuple)):
            cell_id = list(
                set(
                    [
                        "_".join(
                            k.split("_")[key_index_bounds[0] : key_index_bounds[1]]
                        )
                        for k in keys
                    ]
                )
            )[0]
        elif isinstance(keys, str):
            cell_id = keys
    new_frame = pd.concat(frames, axis=1)

    for col in columns:
        number_of_cols = len(new_frame.columns)
        if col in [hdr_norm_cycle, hdr_cum_charge] and normalize_cycles:
            if number_of_cols > 1:
                avg_frame = (
                    new_frame[col].agg(["mean"], axis=1).rename(columns={"mean": col})
                )
            else:
                avg_frame = new_frame[col].copy()

        else:

            new_col_name_mean = col + "_mean"
            new_col_name_std = col + "_std"

            if number_of_cols > 1:
                # very slow:
                avg_frame = (
                    new_frame[col]
                    .agg(["mean", "std"], axis=1)
                    .rename(
                        columns={"mean": new_col_name_mean, "std": new_col_name_std}
                    )
                )
            else:
                avg_frame = pd.DataFrame(
                    data=new_frame[col].values, columns=[new_col_name_mean]
                )
                avg_frame[new_col_name_std] = not_a_number

        new_frames.append(avg_frame)
    final_frame = pd.concat(new_frames, axis=1)

    return final_frame, cell_id


def update_journal_cellpy_data_dir(
    pages, new_path=None, from_path="PureWindowsPath", to_path="Path"
):
    """Update the path in the pages (batch) from one type of OS to another.

    I use this function when I switch from my work PC (windows) to my home
    computer (mac).

    Args:
        pages: the (batch.experiment.)journal.pages object (pandas.DataFrame)
        new_path: the base path (uses prms.Paths.cellpydatadir if not given)
        from_path: type of path to convert from.
        to_path: type of path to convert to.

    Returns:
        journal.pages (pandas.DataFrame)

    """
    # TODO: move this to batch?

    if new_path is None:
        new_path = prms.Paths.cellpydatadir

    from_path = getattr(pathlib, from_path)
    to_path = getattr(pathlib, to_path)

    pages.cellpy_file_names = pages.cellpy_file_names.apply(from_path)
    pages.cellpy_file_names = pages.cellpy_file_names.apply(
        lambda x: to_path(new_path) / x.name
    )
    return pages


def make_new_cell():
    """create an empty CellpyData object."""
    warnings.warn(
        "make_new_cell is deprecated, use CellpyData.vacant instead", DeprecationWarning
    )
    new_cell = cellpy.cellreader.CellpyData(initialize=True)
    return new_cell


def split_experiment(cell, base_cycles=None):
    """Split experiment (CellpyData object) into several sub-experiments.

    Args:
        cell (CellpyData): original cell
        base_cycles (int or list of ints): cycle(s) to do the split on.

    Returns:
        List of CellpyData objects
    """
    warnings.warn(
        "split_experiment is deprecated, use CellpyData.split_many instead",
        DeprecationWarning,
    )

    if base_cycles is None:
        all_cycles = cell.get_cycle_numbers()
        base_cycles = int(np.median(all_cycles))

    cells = list()
    if not isinstance(base_cycles, (list, tuple)):
        base_cycles = [base_cycles]

    dataset = cell.cell
    steptable = dataset.steps
    data = dataset.raw
    summary = dataset.summary

    for b_cycle in base_cycles:
        steptable0, steptable = [
            steptable[steptable.cycle < b_cycle],
            steptable[steptable.cycle >= b_cycle],
        ]
        data0, data = [
            data[data.cycle_index < b_cycle],
            data[data.cycle_index >= b_cycle],
        ]
        summary0, summary = [
            summary[summary.index < b_cycle],
            summary[summary.index >= b_cycle],
        ]

        new_cell = CellpyData.vacant()

        new_cell.cell.steps = steptable0

        new_cell.cell.raw = data0
        new_cell.cell.summary = summary0

        old_cell = CellpyData.vacant()
        old_cell.cell.steps = steptable

        old_cell.cell.raw = data
        old_cell.cell.summary = summary

        for attr in ATTRS_DATASET:
            value = getattr(cell.cell, attr)
            setattr(new_cell.cell, attr, value)
            setattr(old_cell.cell, attr, value)

        for attr in ATTRS_CELLPYDATA:
            value = getattr(cell, attr)
            setattr(new_cell, attr, value)
            setattr(old_cell, attr, value)

        cells.append(new_cell)

    cells.append(old_cell)

    return cells


def add_normalized_cycle_index(summary, nom_cap=None, column_name=None):
    """Adds normalized cycles to the summary data frame.

    This functionality is now also implemented as default when creating
    the summary (make_summary). However it is kept here if you would like to
    redo the normalization, for example if you want to use another nominal
    capacity or if you would like to have more than one normalized cycle index.

    Args:
        summary (pandas.DataFrame): cell summary
        cell (CellpyData): cell object
        nom_cap (float): nominal capacity to use when normalizing. Defaults to
            the nominal capacity defined in the cell object (this is typically
            set during creation of the CellpyData object based on the value
            given in the parameter file).
        column_name (str): name of the new column. Uses the name defined in
            cellpy.parameters.internal_settings as default.

    Returns:
        cell object now with normalized cycle index in its summary.
    """
    hdr_norm_cycle = hdr_summary["normalized_cycle_index"]
    hdr_cum_charge = hdr_summary["cumulated_charge_capacity"]

    if column_name is None:
        column_name = hdr_norm_cycle
    hdr_cum_charge = hdr_cum_charge

    # if nom_cap is None:
    #   nom_cap = cell.cell.nom_cap
    summary[column_name] = summary[hdr_cum_charge] / nom_cap
    return summary


def add_c_rate(cell, nom_cap=None, column_name=None):
    """Adds C-rates to the step table data frame.

    This functionality is now also implemented as default when creating
    the step_table (make_step_table). However it is kept here if you would
    like to recalculate the C-rates, for example if you want to use another
    nominal capacity or if you would like to have more than one column with
    C-rates.

    Args:
        cell (CellpyData): cell object
        nom_cap (float): nominal capacity to use for estimating C-rates.
            Defaults to the nominal capacity defined in the cell object
            (this is typically set during creation of the CellpyData object
            based on the value given in the parameter file).
        column_name (str): name of the new column. Uses the name defined in
            cellpy.parameters.internal_settings as default.

    Returns:
        cell object.
    """

    # now also included in step_table
    # TODO: remove this function
    if column_name is None:
        column_name = hdr_steps["rate_avr"]
    if nom_cap is None:
        nom_cap = cell.cell.nom_cap

    spec_conv_factor = cell.get_converter_to_specific()
    cell.cell.steps[column_name] = abs(
        round(cell.cell.steps.current_avr / (nom_cap / spec_conv_factor), 2)
    )

    return cell


def add_areal_capacity(cell, cell_id, journal):
    """Adds areal capacity to the summary."""

    loading = journal.pages.loc[cell_id, hdr_journal["loading"]]

    cell.cell.summary[hdr_summary["areal_charge_capacity"]] = (
        cell.cell.summary[hdr_summary["charge_capacity"]] * loading / 1000
    )
    cell.cell.summary[hdr_summary["areal_discharge_capacity"]] = (
        cell.cell.summary[hdr_summary["discharge_capacity"]] * loading / 1000
    )
    return cell


def _remove_outliers_from_summary(s, filter_vals, freeze_indexes=None):
    if freeze_indexes is not None:
        try:
            filter_vals[freeze_indexes] = True
        except IndexError:
            logging.critical(
                f"Could not freeze - missing cycle indexes {freeze_indexes}"
            )

    return s[filter_vals]


def remove_outliers_from_summary_on_window(
    s, window_size=3, cut=0.1, iterations=1, col_name=None, freeze_indexes=None
):
    """Removes outliers based based on neighbours"""
    if col_name is None:
        col = hdr_summary["charge_capacity"]

    else:
        col = hdr_summary[col_name]

    def fractional_std(x):
        return np.std(x) / np.mean(x)

    for j in range(iterations):
        fractional_deviation_series = (
            s[col]
            .rolling(window=window_size, center=True, min_periods=1)
            .apply(fractional_std)
        )
        filter_vals = fractional_deviation_series < cut
        s = s[filter_vals]

    s = _remove_outliers_from_summary(s, filter_vals, freeze_indexes=freeze_indexes)

    return s


def remove_outliers_from_summary_on_nn_distance(
    s, distance=0.7, filter_cols=None, freeze_indexes=None
):
    """Remove outliers with missing neighbours.

    Args:
        s (pandas.DataFrame): summary frame
        distance (float): cut-off (all cycles that have a closest neighbour further apart this number will be removed)
        filter_cols (list): list of column headers to perform the filtering on (defaults to charge and discharge capacity)
        freeze_indexes (list): list of cycle indexes that should never be removed (defaults to cycle 1)

    Returns:
        filtered summary (pandas.DataFrame)

    Returns:

    """
    if filter_cols is None:
        filter_cols = [
            hdr_summary["charge_capacity"],
            hdr_summary["discharge_capacity"],
        ]

    def neighbour_window(y):
        y = y.values
        if len(y) == 1:
            # only included in case the pandas rolling function changes in the future
            return 0.5
        if len(y) == 2:
            return abs(np.diff(y)) / np.mean(y)
        else:
            return min(abs(y[1] - y[0]), abs(y[1] - y[2])) / min(
                np.mean(y[0:1]), np.mean(y[1:])
            )

    s2 = s[filter_cols].copy()

    r = s2[filter_cols].rolling(3, center=True, min_periods=1).apply(neighbour_window)
    filter_vals = (r < distance).all(axis=1)

    s = _remove_outliers_from_summary(s, filter_vals, freeze_indexes=freeze_indexes)

    return s


def remove_outliers_from_summary_on_zscore(
    s, zscore_limit=4, filter_cols=None, freeze_indexes=None
):
    """Remove outliers based on z-score.

    Args:
        s (pandas.DataFrame): summary frame
        zscore_limit (int): remove outliers outside this z-score limit
        filter_cols (list): list of column headers to perform the filtering on (defaults to charge and discharge capacity)
        freeze_indexes (list): list of cycle indexes that should never be removed (defaults to cycle 1)

    Returns:
        filtered summary (pandas.DataFrame)
    """

    if freeze_indexes is None:
        freeze_indexes = [1]

    if filter_cols is None:
        filter_cols = [
            hdr_summary["charge_capacity"],
            hdr_summary["discharge_capacity"],
        ]

    s2 = s[filter_cols].copy()

    filter_vals = (np.abs(stats.zscore(s2)) < zscore_limit).all(axis=1)

    s = _remove_outliers_from_summary(s, filter_vals, freeze_indexes=freeze_indexes)

    return s


def remove_outliers_from_summary_on_value(
    s, low=0.0, high=7_000, filter_cols=None, freeze_indexes=None
):
    """Remove outliers based highest and lowest allowed value

    Args:
        s (pandas.DataFrame): summary frame
        low (float): low cut-off (all cycles with values below this number will be removed)
        high (float): high cut-off (all cycles with values above this number will be removed)
        filter_cols (list): list of column headers to perform the filtering on (defaults to charge and discharge capacity)
        freeze_indexes (list): list of cycle indexes that should never be removed (defaults to cycle 1)

    Returns:
        filtered summary (pandas.DataFrame)

    Returns:

    """
    if filter_cols is None:
        filter_cols = [
            hdr_summary["charge_capacity"],
            hdr_summary["discharge_capacity"],
        ]

    s2 = s[filter_cols].copy()

    filter_vals = ((s2[filter_cols] > low) & (s2[filter_cols] < high)).all(axis=1)

    s = _remove_outliers_from_summary(s, filter_vals, freeze_indexes=freeze_indexes)

    return s


def remove_outliers_from_summary_on_index(s, indexes=None, remove_last=False):
    """Remove rows with supplied indexes (where the indexes typically are cycle-indexes).

    Args:
        s (pandas.DataFrame): cellpy summary to process
        indexes (list): list of indexes
        remove_last (bool): remove the last point

    Returns:
        pandas.DataFrame
    """
    logging.debug("removing outliers from summary on index")
    if indexes is None:
        indexes = []

    selection = s.index.isin(indexes)
    if remove_last:
        selection[-1] = True
    return s[~selection]


def remove_last_cycles_from_summary(s, last=None):
    """Remove last rows after given cycle number"""

    if last is not None:
        s = s.loc[s.index <= last, :]
    return s


def remove_first_cycles_from_summary(s, first=None):
    """Remove last rows after given cycle number"""

    if first is not None:
        s = s.loc[s.index >= first, :]
    return s


def yank_after(b, last=None, keep_old=False):
    """Cut all cycles after a given cycle index number.

    Args:
        b (batch object): the batch object to perform the cut on.
        last (int or dict {cell_name: last index}): the last cycle index to keep
            (if dict: use individual last indexes for each cell).
        keep_old (bool): keep the original batch object and return a copy instead.

    Returns:
        batch object if keep_old is True, else None
    """

    if keep_old:
        b = deepcopy(b)

    if last is None:
        return b

    for cell_number, cell_label in enumerate(b.experiment.cell_names):
        c = b.experiment.data[cell_label]
        s = c.cell.summary
        if isinstance(last, dict):
            last_this_cell = last.get(cell_label, None)
        else:
            last_this_cell = last
        s = remove_last_cycles_from_summary(s, last_this_cell)
        c.cell.summary = s
    if keep_old:
        return b


def yank_before(b, first=None, keep_old=False):
    """Cut all cycles before a given cycle index number.

    Args:
        b (batch object): the batch object to perform the cut on.
        first (int or dict {cell_name: first index}): the first cycle index to keep
            (if dict: use individual first indexes for each cell).
        keep_old (bool): keep the original batch object and return a copy instead.

    Returns:
        batch object if keep_old is True, else None
    """

    if keep_old:
        b = deepcopy(b)

    if first is None:
        return b

    for cell_number, cell_label in enumerate(b.experiment.cell_names):
        c = b.experiment.data[cell_label]
        s = c.cell.summary
        if isinstance(first, dict):
            first_this_cell = first.get(cell_label, None)
        else:
            first_this_cell = first
        s = remove_first_cycles_from_summary(s, first_this_cell)
        c.cell.summary = s
    if keep_old:
        return b


def yank_outliers(
    b,
    zscore_limit=None,
    low=0.0,
    high=7_000.0,
    filter_cols=None,
    freeze_indexes=None,
    remove_indexes=None,
    remove_last=False,
    iterations=1,
    zscore_multiplyer=1.3,
    distance=None,
    window_size=None,
    window_cut=0.1,
    keep_old=False,
):
    """Remove outliers from a batch object.

    Args:
        b (cellpy.utils.batch object): the batch object to perform filtering one (required).
        zscore_limit (int): will filter based on z-score if given.
        low (float): low cut-off (all cycles with values below this number will be removed)
        high (float): high cut-off (all cycles with values above this number will be removed)
        filter_cols (str): what columns to filter on.
        freeze_indexes (list): indexes (cycles) that should never be removed.
        remove_indexes (dict or list): if dict, look-up on cell label, else a list that will be the same for all
        remove_last (dict or bool): if dict, look-up on cell label.
        iterations (int): repeat z-score filtering if `zscore_limit` is given.
        zscore_multiplyer (int): multiply `zscore_limit` with this number between each z-score filtering
            (should usually be less than 1).
        distance (float): nearest neighbour normalised distance required (typically 0.5).
        window_size (int): number of cycles to include in the window.
        window_cut (float): cut-off.

        keep_old (bool): perform filtering of a copy of the batch object
            (not recommended at the moment since it then loads the full cellpyfile).

    Returns:
        if keep_old: new cellpy.utils.batch object.
        else: dictionary of removed cycles
    """

    if keep_old:
        b = deepcopy(b)

    removed_cycles = dict()

    # remove based on indexes and values
    for cell_number, cell_label in enumerate(b.experiment.cell_names):
        logging.debug(f"yanking {cell_label} ")
        c = b.experiment.data[cell_label]
        s = c.cell.summary
        before = set(s.index)
        if remove_indexes is not None:
            logging.debug("removing indexes")
            if isinstance(remove_indexes, dict):
                remove_indexes_this_cell = remove_indexes.get(cell_label, None)
            else:
                remove_indexes_this_cell = remove_indexes

            if isinstance(remove_last, dict):
                remove_last_this_cell = remove_last.get(cell_label, None)
            else:
                remove_last_this_cell = remove_last

            s = remove_outliers_from_summary_on_index(
                s, remove_indexes_this_cell, remove_last_this_cell
            )

        s = remove_outliers_from_summary_on_value(
            s,
            low=low,
            high=high,
            filter_cols=filter_cols,
            freeze_indexes=freeze_indexes,
        )

        if distance is not None:
            s = remove_outliers_from_summary_on_nn_distance(
                s,
                distance=distance,
                filter_cols=filter_cols,
                freeze_indexes=freeze_indexes,
            )
            c.cell.summary = s

        if window_size is not None:
            s = remove_outliers_from_summary_on_window(
                s,
                window_size=window_size,
                cut=window_cut,
                iterations=iterations,
                freeze_indexes=freeze_indexes,
            )

        removed = before - set(s.index)
        c.cell.summary = s
        if removed:
            removed_cycles[cell_label] = list(removed)

    if zscore_limit is not None:
        logging.info("using the zscore - removed cycles not kept track on")
        for j in range(iterations):
            tot_rows_removed = 0
            for cell_number, cell_label in enumerate(b.experiment.cell_names):
                c = b.experiment.data[cell_label]
                n1 = len(c.cell.summary)
                s = remove_outliers_from_summary_on_zscore(
                    c.cell.summary,
                    filter_cols=filter_cols,
                    zscore_limit=zscore_limit,
                    freeze_indexes=freeze_indexes,
                )
                # TODO: populate removed_cycles
                rows_removed = n1 - len(s)
                tot_rows_removed += rows_removed
                c.cell.summary = s
            if tot_rows_removed == 0:
                break
            zscore_limit *= zscore_multiplyer

    if keep_old:
        return b
    else:
        return removed_cycles


# from helpers - updated
def concatenate_summaries(
    b,
    max_cycle=None,
    rate=None,
    on="charge",
    columns=None,
    column_names=None,
    normalize_capacity_on=None,
    scale_by=None,
    nom_cap=None,
    normalize_cycles=False,
    add_areal=False,
    group_it=False,
    rate_std=None,
    rate_column=None,
    inverse=False,
    inverted=False,
    key_index_bounds=[1, -2],
    melt=False,
    cell_type_split_position="auto",
):

    """Merge all summaries in a batch into a gigantic summary data frame.

    TODO: Allow also dictionaries of cell objects.
    TODO: Allow iterating through batch-objects (for id, name in b.iteritems() or similar)

    Args:
        b (cellpy.batch object): the batch with the cells.
        max_cycle (int): drop all cycles above this value.
        rate (float): filter on rate (C-rate)
        on (str or list of str): only select cycles if based on the rate of this step-type (e.g. on="charge").
        columns (list): selected column(s) (using cellpy name) [defaults to "charge_capacity"]
        column_names (list): selected column(s) (using exact column name)
        normalize_capacity_on (list): list of cycle numbers that will be used for setting the basis of the normalization (typically the first few cycles after formation)
        scale_by (float or str): scale the normalized data with nominal capacity if "nom_cap", or given value (defaults to one).
        nom_cap (float): nominal capacity of the cell
        normalize_cycles (bool): perform a normalisation of the cycle numbers (also called equivalent cycle index)
        add_areal (bool):  add areal capacity to the summary
        group_it (bool): if True, average pr group.
        rate_std (float): allow for this inaccuracy when selecting cycles based on rate
        rate_column (str): name of the column containing the C-rates.
        inverse (bool): select steps that does not have the given C-rate.
        inverted (bool): select cycles that does not have the steps filtered by given C-rate.
        key_index_bounds (list): used when creating a common label for the cells by splitting and combining from
            key_index_bound[0] to key_index_bound[1].
        melt (bool): return frame as melted (long format).
        cell_type_split_position (int | None | "auto"): list item number for creating a cell type identifier
            after performing a split("_") on the cell names (only valid if melt==True). Set to None to not
            include a cell_type col.

    Returns:
        Multi-index ``pandas.DataFrame``

    Notes:
        The returned ``DataFrame`` has the following structure:

        - top-level columns or second col for melted: cell-names (cell_name)
        - second-level columns or first col for melted: summary headers (summary_headers)
        - row-index or third col for melted: cycle number (cycle_index)
        - cell_type on forth col for melted if cell_type_split_position is given

    """

    # TODO: refactor me
    # TODO: check if selecting normalize_cycles and group_it performs the operation in logical order
    if normalize_capacity_on is not None:
        default_columns = ["normalized_charge_capacity"]
    else:
        default_columns = ["charge_capacity"]

    import logging

    reserved_cell_label_names = ["FC"]

    hdr_norm_cycle = hdr_summary["normalized_cycle_index"]
    hdr_cum_charge = hdr_summary["cumulated_charge_capacity"]

    cell_names_nest = []
    frames = []
    keys = []

    if columns is not None:
        if any(s.startswith("areal_") for s in columns):
            add_areal = True
        if normalize_capacity_on is not None:
            _columns = []
            for name in columns:
                if name.startswith("normalized_"):
                    _columns.append(hdr_summary[name])
                else:
                    _columns.append(hdr_summary["normalized_" + name])
            columns = _columns
        else:
            columns = [hdr_summary[name] for name in columns]
    else:
        columns = []

    if column_names is not None:
        columns += column_names

    if len(columns) > 1:
        columns = [hdr_summary[name] for name in default_columns]

    logging.info(f"concatenating the following columns: {columns}")
    normalize_cycles_headers = []

    if normalize_cycles:
        if hdr_norm_cycle not in columns:
            normalize_cycles_headers.append(hdr_norm_cycle)
        if hdr_cum_charge not in columns:
            normalize_cycles_headers.append(hdr_cum_charge)

    if group_it:
        g = b.pages.groupby("group")
        for gno, b_sub in g:
            cell_names_nest.append(list(b_sub.index))
    else:
        cell_names_nest.append(list(b.experiment.cell_names))

    for cell_names in cell_names_nest:
        frames_sub = []
        keys_sub = []

        for cell_id in cell_names:
            logging.debug(f"Processing [{cell_id}]")
            c = b.experiment.data[cell_id]

            if not c.empty:
                if max_cycle is not None:
                    c = c.drop_from(max_cycle + 1)
                if add_areal:
                    c = add_areal_capacity(c, cell_id, b.experiment.journal)

                if normalize_capacity_on is not None:
                    if scale_by == "nom_cap":
                        if nom_cap is None:
                            scale_by = c.cell.nom_cap
                        else:
                            scale_by = nom_cap
                    elif scale_by is None:
                        scale_by = 1.0

                    c = add_normalized_capacity(
                        c, norm_cycles=normalize_capacity_on, scale=scale_by
                    )

                if rate is not None:
                    s = select_summary_based_on_rate(
                        c,
                        rate=rate,
                        on=on,
                        rate_std=rate_std,
                        rate_column=rate_column,
                        inverse=inverse,
                        inverted=inverted,
                    )

                else:
                    s = c.cell.summary

                if columns is not None:
                    if normalize_cycles:
                        s = s.loc[:, normalize_cycles_headers + columns].copy()
                    else:
                        s = s.loc[:, columns].copy()

                if normalize_cycles:
                    if nom_cap is None:
                        _nom_cap = c.cell.nom_cap
                    else:
                        _nom_cap = nom_cap

                    if not group_it:
                        s = add_normalized_cycle_index(s, nom_cap=_nom_cap)
                        if hdr_cum_charge not in columns:
                            s = s.drop(columns=hdr_cum_charge)

                    s = s.reset_index(drop=True)

                frames_sub.append(s)
                keys_sub.append(cell_id)

        if group_it:
            try:
                if normalize_cycles:
                    s, cell_id = _make_average(
                        frames_sub,
                        keys_sub,
                        normalize_cycles_headers + columns,
                        True,
                        key_index_bounds,
                    )
                    s = add_normalized_cycle_index(s, nom_cap=_nom_cap)
                    if hdr_cum_charge not in columns:
                        s = s.drop(columns=hdr_cum_charge)
                else:
                    s, cell_id = _make_average(
                        frames_sub, keys_sub, columns, key_index_bounds=key_index_bounds
                    )
            except ValueError as e:
                print("could not make average!")
                print(e)
            else:
                frames.append(s)
                keys.append(cell_id)
        else:
            frames.extend(frames_sub)
            keys.extend(keys_sub)

    if frames:
        if len(set(keys)) != len(keys):
            logging.info("Got several columns with same test-name")
            logging.info("Renaming.")
            used_names = []
            new_keys = []
            for name in keys:
                while True:
                    if name in used_names:
                        name += "x"
                    else:
                        break
                new_keys.append(name)
                used_names.append(name)
            keys = new_keys
        cdf = pd.concat(frames, keys=keys, axis=1)
        cdf = cdf.rename_axis(columns=["cell_name", "summary_header"])
        if melt:
            cdf = cdf.reset_index(drop=False).melt(
                id_vars=hdr_summary.cycle_index, value_name="value"
            )
            melted_column_order = [
                "summary_header",
                "cell_name",
                hdr_summary.cycle_index,
                "value",
            ]

            if cell_type_split_position == "auto":
                cell_type_split_position = 1
                _pp = cdf.cell_name.str.split("_", expand=True).values[0]
                for _p in _pp[1:]:
                    if _p not in reserved_cell_label_names:
                        break
                    cell_type_split_position += 1

            if cell_type_split_position is not None:

                cdf = cdf.assign(
                    cell_type=cdf.cell_name.str.split("_", expand=True)[
                        cell_type_split_position
                    ]
                )
                melted_column_order.insert(-2, "cell_type")
            cdf = cdf.reindex(columns=melted_column_order)

        return cdf
    else:
        logging.info("Empty - nothing to concatenate!")
        return pd.DataFrame()


def create_rate_column(df, nom_cap, spec_conv_factor, column="current_avr"):
    """Adds a rate column to the dataframe (steps)."""

    col = abs(round(df[column] / (nom_cap / spec_conv_factor), 2))
    return col


def select_summary_based_on_rate(
    cell,
    rate=None,
    on=None,
    rate_std=None,
    rate_column=None,
    inverse=False,
    inverted=False,
    fix_index=True,
):
    """Select only cycles charged or discharged with a given rate.

    Parameters:
        cell (cellpy.CellpyData)
        rate (float): the rate to filter on. Remark that it should be given
            as a float, i.e. you will have to convert from C-rate to
            the actual numeric value. For example, use rate=0.05 if you want
            to filter on cycles that has a C/20 rate.
        on (str): only select cycles if based on the rate of this step-type (e.g. on="charge").
        rate_std (float): allow for this inaccuracy in C-rate when selecting cycles
        rate_column (str): column header name of the rate column,
        inverse (bool): select steps that do not have the given C-rate.
        inverted (bool): select cycles that do not have the steps filtered by given C-rate.
        fix_index (bool): automatically set cycle indexes as the index for the summary dataframe if not already set.

    Returns:
        filtered summary (Pandas.DataFrame).
    """

    if on is None:
        on = ["charge"]
    else:
        if not isinstance(on, (list, tuple)):
            on = [on]

    if rate_column is None:
        rate_column = hdr_steps["rate_avr"]

    if on:
        on_column = hdr_steps["type"]

    if rate is None:
        rate = 0.05

    if rate_std is None:
        rate_std = 0.1 * rate

    cycle_number_header = hdr_summary["cycle_index"]

    step_table = cell.cell.steps
    summary = cell.cell.summary

    if summary.index.name != cycle_number_header:
        warnings.warn(
            f"{cycle_number_header} not set as index\n"
            f"Current index :: {summary.index}\n"
        )

        if fix_index:
            summary.set_index(cycle_number_header, drop=True, inplace=True)
        else:
            print(f"{cycle_number_header} not set as index!")
            print(f"Please, set the cycle index header as index before proceeding!")
            return summary

    if on:
        cycles_mask = (
            (step_table[rate_column] < (rate + rate_std))
            & (step_table[rate_column] > (rate - rate_std))
            & (step_table[on_column].isin(on))
        )
    else:
        cycles_mask = (step_table[rate_column] < (rate + rate_std)) & (
            step_table[rate_column] > (rate - rate_std)
        )

    if inverse:
        cycles_mask = ~cycles_mask

    filtered_step_table = step_table[cycles_mask]
    filtered_cycles = filtered_step_table[hdr_steps.cycle].unique()

    if inverted:
        filtered_index = summary.index.difference(filtered_cycles)
    else:
        filtered_index = summary.index.intersection(filtered_cycles)

    if filtered_index.empty:
        warnings.warn("EMPTY")

    return summary.loc[filtered_index, :]


def add_normalized_capacity(
    cell, norm_cycles=None, individual_normalization=False, scale=1.0
):
    """Add normalized capacity to the summary.

    Args:
        cell (CellpyData): cell to add normalized capacity to.
        norm_cycles (list of ints): the cycles that will be used to find
            the normalization factor from (averaging their capacity)
        individual_normalization (bool): find normalization factor for both
            the charge and the discharge if true, else use normalization factor
            from charge on both charge and discharge.
        scale (float): scale of normalization (default is 1.0).

    Returns:
        cell (CellpyData) with added normalization capacity columns in
        the summary.
    """

    if norm_cycles is None:
        norm_cycles = [1, 2, 3, 4, 5]

    col_name_charge = hdr_summary["charge_capacity"]
    col_name_discharge = hdr_summary["discharge_capacity"]
    col_name_norm_charge = hdr_summary["normalized_charge_capacity"]
    col_name_norm_discharge = hdr_summary["normalized_discharge_capacity"]

    try:
        norm_val_charge = cell.cell.summary.loc[norm_cycles, col_name_charge].mean()
    except KeyError as e:
        print(f"Oh no! Are you sure these cycle indexes exist?")
        print(f"  norm_cycles: {norm_cycles}")
        print(f"  cycle indexes: {list(cell.cell.summary.index)}")
        raise KeyError from e
    if individual_normalization:
        norm_val_discharge = cell.cell.summary.loc[
            norm_cycles, col_name_discharge
        ].mean()
    else:
        norm_val_discharge = norm_val_charge

    for col_name, norm_col_name, norm_value in zip(
        [col_name_charge, col_name_discharge],
        [col_name_norm_charge, col_name_norm_discharge],
        [norm_val_charge, norm_val_discharge],
    ):
        cell.cell.summary[norm_col_name] = (
            scale * cell.cell.summary[col_name] / norm_value
        )

    return cell


def load_and_save_resfile(filename, outfile=None, outdir=None, mass=1.00):
    """Load a raw data file and save it as cellpy-file.

    Args:
        mass (float): active material mass [mg].
        outdir (path): optional, path to directory for saving the hdf5-file.
        outfile (str): optional, name of hdf5-file.
        filename (str): name of the resfile.

    Returns:
        out_file_name (str): name of saved file.
    """
    warnings.warn(DeprecationWarning("This option will be removed in v.0.4.0"))
    d = CellpyData()

    if not outdir:
        outdir = prms.Paths.cellpydatadir

    if not outfile:
        outfile = os.path.basename(filename).split(".")[0] + ".h5"
        outfile = os.path.join(outdir, outfile)

    print("filename:", filename)
    print("outfile:", outfile)
    print("outdir:", outdir)
    print("mass:", mass, "mg")

    d.from_raw(filename)
    d.set_mass(mass)
    d.make_step_table()
    d.make_summary()
    d.save(filename=outfile)
    d.to_csv(datadir=outdir, cycles=True, raw=True, summary=True)
    return outfile
